# same_interface_test.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""dbapi, dptapi, sqlt3api, and apswapi must present the same interface"""

import unittest
from copy import copy, deepcopy
import os

try:
    from .. import dbapi
    bsddb = True
except:
    bsddb = False
try:
    from .. import sqlite3api
    sqlite3 = True
except:
    sqlite3 = False
try:
    from .. import apswapi
    apsw = True
except:
    apsw = False
try:
    from .. import dptbase
    dpt = True
except:
    dpt = False


class SameInterfaceTC(unittest.TestCase):

    def setUp(self):
        self.bsddb = bsddb
        self.sqlite3 = sqlite3
        self.apsw = apsw
        self.dpt = dpt

    def tearDown(self):
        pass

    def test_interface_01(self):
        api = {}
        if self.bsddb:
            api[bsddb] = {a for a in dir(dbapi.DBbitapi)
                          if not a.startswith('__') and not a.endswith('__')}
        if self.sqlite3:
            api[sqlite3] = {a for a in dir(sqlite3api.Sqlite3bitapi)
                            if not a.startswith('__') and not a.endswith('__')}
        if self.apsw:
            api[apsw] = {a for a in dir(apswapi.Sqlite3bitapi)
                         if not a.startswith('__') and not a.endswith('__')}
        if self.dpt:
            api[dpt] = {a for a in dir(dptbase.DPTbase)
                        if not a.startswith('__') and not a.endswith('__')}
        for api_a, api_b in ((self.bsddb, self.sqlite3),
                             (self.bsddb, self.apsw),
                             (self.bsddb, self.dpt),
                             (self.sqlite3, self.apsw),
                             (self.sqlite3, self.dpt),
                             (self.apsw, self.dpt),
                             ):
            if api_a and api_b:
                self.assertEqual(api[api_a], api[api_b])
        self.assertNotEqual(len(api), 0)
        for v in api.values():
            self.assertEqual(len(v), 50)


if __name__ == '__main__':
    runner = unittest.TextTestRunner
    loader = unittest.defaultTestLoader.loadTestsFromTestCase
    runner().run(loader(SameInterfaceTC))

