# where_test.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""where tests"""

import unittest
from copy import copy, deepcopy
import re

from .. import where


class WhereTC(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test___raises(self):
        """"""
        pass

    def test___copy(self):
        """"""
        pass

    def test____assumptions(self):
        """"""
        msg = 'Failure of this test invalidates all other tests'

    def test____module_constants(self):
        """"""
        constants = ((where.DOUBLE_QUOTE_STRING, '".*?"'),
                     (where.SINGLE_QUOTE_STRING, "'.*?'"),
                     (where.LEFT_PARENTHESIS, '('),
                     (where.RIGHT_PARENTHESIS, ')'),
                     (where.OR, 'or'),
                     (where.TO, 'to'),
                     (where.IS, 'is'),
                     (where.EQ, 'eq'),
                     (where.NE, 'ne'),
                     (where.LT, 'lt'),
                     (where.LE, 'le'),
                     (where.GT, 'gt'),
                     (where.GE, 'ge'),
                     (where.NOT, 'not'),
                     (where.NOR, 'nor'),
                     (where.AND, 'and'),
                     (where.NUM, 'num'),
                     (where.LIKE, 'like'),
                     (where.FROM, 'from'),
                     (where.ALPHA, 'alpha'),
                     (where.ABOVE, 'above'),
                     (where.AFTER, 'after'),
                     (where.BELOW, 'below'),
                     (where.BEFORE, 'before'),
                     (where.PRESENT, 'present'),
                     (where.STRING, '.*?'),
                     (where.LEADING_SPACE, '(?<=\s)'),
                     (where.TRAILING_SPACE, '(?=\s)'),
                     )
        for a, v in constants:
            self.assertEqual(a, v)
        self.assertEqual(where.KEYWORDS,
                         frozenset((where.LEFT_PARENTHESIS,
                                    where.RIGHT_PARENTHESIS,
                                    where.OR,
                                    where.TO,
                                    where.IS,
                                    where.EQ,
                                    where.NE,
                                    where.LT,
                                    where.LE,
                                    where.GT,
                                    where.GE,
                                    where.NOT,
                                    where.NOR,
                                    where.AND,
                                    where.NUM,
                                    where.LIKE,
                                    where.FROM,
                                    where.ALPHA,
                                    where.ABOVE,
                                    where.AFTER,
                                    where.BELOW,
                                    where.BEFORE,
                                    where.PRESENT,
                                    )))
        self.assertEqual(
            {where.LEADING_SPACE,
             where.TRAILING_SPACE,
             where.STRING,
             where.DOUBLE_QUOTE_STRING,
             where.SINGLE_QUOTE_STRING,}.union(where.KEYWORDS),
            {a for a, b in constants})
        self.assertEqual(where.WHERE_RE.__class__, re.compile('').__class__)
        self.assertEqual(
            where.WHERE_RE.flags & (re.IGNORECASE|re.DOTALL),
            re.IGNORECASE|re.DOTALL)
        self.assertEqual(
            where.WHERE_RE.pattern,
            '|'.join(
                (where.DOUBLE_QUOTE_STRING,
                 where.SINGLE_QUOTE_STRING,
                 ''.join(('\\', where.LEFT_PARENTHESIS)),
                 ''.join(('\\', where.RIGHT_PARENTHESIS)),
                 where.OR.join(('(?<=\s|\))', '(?=\s|\()')),
                 where.TO.join((where.LEADING_SPACE, where.TRAILING_SPACE)),
                 where.IS.join((where.LEADING_SPACE, where.TRAILING_SPACE)),
                 where.EQ.join((where.LEADING_SPACE, where.TRAILING_SPACE)),
                 where.NE.join((where.LEADING_SPACE, where.TRAILING_SPACE)),
                 where.LT.join((where.LEADING_SPACE, where.TRAILING_SPACE)),
                 where.LE.join((where.LEADING_SPACE, where.TRAILING_SPACE)),
                 where.GT.join((where.LEADING_SPACE, where.TRAILING_SPACE)),
                 where.GE.join((where.LEADING_SPACE, where.TRAILING_SPACE)),
                 where.NOT.join(('\A', '(?=\s|\()')),
                 where.NOT.join(('(?<=\s|\()', '(?=\s|\()')),
                 where.NOR.join(('(?<=\s|\()', '(?=\s|\()')),
                 where.AND.join(('(?<=\s|\))', '(?=\s|\()')),
                 where.NUM.join((where.LEADING_SPACE, where.TRAILING_SPACE)),
                 where.LIKE.join((where.LEADING_SPACE, where.TRAILING_SPACE)),
                 where.FROM.join((where.LEADING_SPACE, where.TRAILING_SPACE)),
                 where.ALPHA.join((where.LEADING_SPACE, where.TRAILING_SPACE)),
                 where.ABOVE.join((where.LEADING_SPACE, where.TRAILING_SPACE)),
                 where.AFTER.join((where.LEADING_SPACE, where.TRAILING_SPACE)),
                 where.BELOW.join((where.LEADING_SPACE, where.TRAILING_SPACE)),
                 where.BEFORE.join((where.LEADING_SPACE, where.TRAILING_SPACE)),
                 where.PRESENT.join((where.LEADING_SPACE, '(?=\s|\)|\Z)')),
                 where.STRING,
                 )).join(('(', ')')))

    def test___init__(self):
        """"""
        w = where.Where('')
        self.assertEqual(len(w.__dict__), 20)
        self.assertEqual(w.statement, '')
        self.assertEqual(w.node, None)
        self.assertEqual(w.tokens, None)
        self.assertEqual(
            w.token_dispatch,
            {where.LEFT_PARENTHESIS: w.left_parenthesis,
             where.RIGHT_PARENTHESIS: w.right_parenthesis,
             where.OR: w.boolean,
             where.TO: w.range_end,
             where.IS: w.is_,
             where.EQ: w.compare,
             where.NE: w.compare,
             where.LT: w.compare,
             where.LE: w.compare,
             where.GT: w.compare,
             where.GE: w.compare,
             where.NOT: w.invert,
             where.NOR: w.boolean,
             where.AND: w.boolean,
             where.NUM: w.alphanumeric,
             where.LIKE: w.like,
             where.FROM: w.range_start,
             where.ALPHA: w.alphanumeric,
             where.ABOVE: w.range_start,
             where.AFTER: w.compare,
             where.BELOW: w.range_end,
             where.BEFORE: w.compare,
             where.PRESENT: w.present,
             None: w.finish,
             False: w.finish,
             })
        self.assertEqual(
            w.present_dispatch,
            {None: w.report,
             False: w.report,
             w.value: w._field_present,
             w.left_parenthesis: w.report,
             w.right_parenthesis: w.report,
             w.boolean: w._boolean_present,
             w.range_end: w.report,
             w.compare: w.report,
             w.is_: w.report,
             w.like: w.report,
             w.invert: w._invert_present,
             w.range_start: w.report,
             w.present: w.report,
             w.alphanumeric: w.report,
             })
        self.assertEqual(
            w.range_start_dispatch,
            {None: w.report,
             False: w.report,
             w.value: w._field_range_start,
             w.left_parenthesis: w.report,
             w.right_parenthesis: w.report,
             w.boolean: w._boolean_range_start,
             w.range_end: w.report,
             w.compare: w.report,
             w.is_: w.report,
             w.like: w.report,
             w.invert: w._invert_range_start,
             w.range_start: w.report,
             w.present: w.report,
             w.alphanumeric: w._alphanumeric_range_start,
             })
        self.assertEqual(
            w.range_end_dispatch,
            {None: w.report,
             False: w.report,
             w.value: w._value_range_end,
             w.left_parenthesis: w.report,
             w.right_parenthesis: w.report,
             w.boolean: w.report,
             w.range_end: w.report,
             w.compare: w.report,
             w.is_: w.report,
             w.like: w.report,
             w.invert: w.report,
             w.range_start: w.report,
             w.present: w.report,
             w.alphanumeric: w.report,
             })
        self.assertEqual(
            w.boolean_dispatch,
            {None: w.report,
             False: w.report,
             w.value: w._value_boolean,
             w.left_parenthesis: w.report,
             w.right_parenthesis: w._right_parenthesis_boolean,
             w.boolean: w.report,
             w.range_end: w.report,
             w.compare: w.report,
             w.is_: w.report,
             w.like: w.report,
             w.invert: w.report,
             w.range_start: w.report,
             w.present: w._present_boolean,
             w.alphanumeric: w.report,
             })
        self.assertEqual(
            w.invert_dispatch,
            {None: w.report,
             False: w._first_token_invert,
             w.value: w._field_invert,
             w.left_parenthesis: w._left_parenthesis_invert,
             w.right_parenthesis: w.report,
             w.boolean: w._boolean_invert,
             w.range_end: w.report,
             w.compare: w.report,
             w.is_: w._is_invert,
             w.like: w.report,
             w.invert: w.report,
             w.range_start: w.report,
             w.present: w.report,
             w.alphanumeric: w.report,
             })
        self.assertEqual(
            w.compare_dispatch,
            {None: w.report,
             False: w.report,
             w.value: w._field_compare,
             w.left_parenthesis: w.report,
             w.right_parenthesis: w.report,
             w.boolean: w._boolean_compare,
             w.range_end: w.report,
             w.compare: w.report,
             w.is_: w.report,
             w.like: w.report,
             w.invert: w._invert_compare,
             w.range_start: w.report,
             w.present: w.report,
             w.alphanumeric: w._alphanumeric_compare,
             })
        self.assertEqual(
            w.is_dispatch,
            {None: w.report,
             False: w.report,
             w.value: w._field_is,
             w.left_parenthesis: w.report,
             w.right_parenthesis: w.report,
             w.boolean: w._boolean_is,
             w.range_end: w.report,
             w.compare: w.report,
             w.is_: w.report,
             w.like: w.report,
             w.invert: w.report,
             w.range_start: w.report,
             w.present: w.report,
             w.alphanumeric: w.report,
             })
        self.assertEqual(
            w.like_dispatch,
            {None: w.report,
             False: w.report,
             w.value: w._field_like,
             w.left_parenthesis: w.report,
             w.right_parenthesis: w.report,
             w.boolean: w._boolean_like,
             w.range_end: w.report,
             w.compare: w.report,
             w.is_: w.report,
             w.like: w.report,
             w.invert: w._invert_like,
             w.range_start: w.report,
             w.present: w.report,
             w.alphanumeric: w.report,
             })
        self.assertEqual(
            w.left_parenthesis_dispatch,
            {None: w.report,
             False: w._first_token_left_parenthesis,
             w.value: w.report,
             w.left_parenthesis: w._left_parenthesis_left_parenthesis,
             w.right_parenthesis: w.report,
             w.boolean: w._boolean_left_parenthesis,
             w.range_end: w.report,
             w.compare: w.report,
             w.is_: w.report,
             w.like: w.report,
             w.invert: w.report,
             w.range_start: w.report,
             w.present: w.report,
             w.alphanumeric: w.report,
             })
        self.assertEqual(
            w.right_parenthesis_dispatch,
            {None: w.report,
             False: w.report,
             w.value: w._value_right_parenthesis,
             w.left_parenthesis: w.report,
             w.right_parenthesis: w._right_parenthesis_right_parenthesis,
             w.boolean: w.report,
             w.range_end: w.report,
             w.compare: w.report,
             w.is_: w.report,
             w.like: w.report,
             w.invert: w.report,
             w.range_start: w.report,
             w.present: w._present_right_parenthesis,
             w.alphanumeric: w.report,
             })
        self.assertEqual(
            w.value_dispatch,
            {None: w.report,
             False: w._first_token_field,
             w.value: w.report,
             w.left_parenthesis: w._left_parenthesis_field,
             w.right_parenthesis: w.report,
             w.boolean: w._value_or_field_after_boolean,
             w.range_end: w._range_end_value,
             w.compare: w._compare_value,
             w.is_: w._is_value,
             w.like: w._like_value,
             w.invert: w._value_or_field_after_invert,
             w.range_start: w._range_start_value,
             w.present: w.report,
             w.alphanumeric: w.report,
             })
        self.assertEqual(
            w.alphanumeric_dispatch,
            {None: w.report,
             False: w.report,
             w.value: w._field_alphanumeric,
             w.left_parenthesis: w.report,
             w.right_parenthesis: w.report,
             w.boolean: w._boolean_alphanumeric,
             w.range_end: w.report,
             w.compare: w.report,
             w.is_: w.report,
             w.like: w.report,
             w.invert: w._invert_alphanumeric,
             w.range_start: w.report,
             w.present: w.report,
             w.alphanumeric: w.report,
             })
        self.assertEqual(
            w.none_dispatch,
            {None: w.report,
             False: w._finish,
             w.value: w._finish,
             w.left_parenthesis: w.report,
             w.right_parenthesis: w._finish,
             w.boolean: w.report,
             w.range_end: w.report,
             w.compare: w.report,
             w.is_: w.report,
             w.like: w.report,
             w.invert: w.report,
             w.range_start: w.report,
             w.present: w._finish,
             w.alphanumeric: w.report,
             })
        self.assertEqual(w._processors, None)
        self.assertEqual(w._f_or_v, None)
        self.assertEqual(w._not, None)


class Where_MethodsTC(unittest.TestCase):

    def setUp(self):
        self.w = where.Where('')
        self.wc = where.WhereClause()
        self.w.node = self.wc
        self.w.node.left = where.WhereClause()

    def tearDown(self):
        del self.w

    def test__first_token_field(self):
        """"""
        w = self.w
        w._first_token_field('')
        self.assertEqual(w.node.field, '')

    def test__field_range_start(self):
        """"""
        w = self.w
        w._field_range_start('')
        self.assertEqual(w.node.condition, '')

    def test__boolean_range_start(self):
        """"""
        w = self.w
        w._boolean_range_start('')
        self.assertEqual(w.node.condition, '')

    def test__invert_range_start(self):
        """"""
        w = self.w
        w._invert_range_start('')
        self.assertEqual(w.node.condition, '')

    def test__alphanumeric_range_start(self):
        """"""
        w = self.w
        w._alphanumeric_range_start('')
        self.assertEqual(w.node.condition, '')

    def test__range_start_value(self):
        """"""
        w = self.w
        w._range_start_value('')
        self.assertEqual(w.node.value, '')

    def test__value_range_end(self):
        """"""
        w = self.w
        w.node.condition = ''
        w._value_range_end('')
        self.assertEqual(w.node.condition, ('', ''))

    def test__range_end_value(self):
        """"""
        w = self.w
        w.node.value = ''
        w._range_end_value('')
        self.assertEqual(w.node.value, ('', ''))

    def test__value_boolean(self):
        """"""
        w = self.w
        w._value_boolean('')
        self.assertEqual(w.node.operator, '')

    def test__present_boolean(self):
        """"""
        w = self.w
        w._present_boolean('')
        self.assertEqual(w.node.operator, '')

    def test__boolean_left_parenthesis(self):
        """"""
        w = self.w
        w._boolean_left_parenthesis('')

    def test__left_parenthesis_field(self):
        """"""
        w = self.w
        w._left_parenthesis_field('')
        self.assertEqual(w.node.field, '')

    def test__field_compare(self):
        """"""
        w = self.w
        w._field_compare('')
        self.assertEqual(w.node.condition, '')

    def test__boolean_compare(self):
        """"""
        w = self.w
        w._boolean_compare('')
        self.assertEqual(w.node.condition, '')

    def test__invert_compare(self):
        """"""
        w = self.w
        w._invert_compare('')
        self.assertEqual(w.node.condition, '')

    def test__alphanumeric_compare(self):
        """"""
        w = self.w
        w._alphanumeric_compare('')
        self.assertEqual(w.node.condition, '')

    def test__field_like(self):
        """"""
        w = self.w
        w._field_like('')
        self.assertEqual(w.node.condition, '')

    def test__field_is(self):
        """"""
        w = self.w
        w._field_is('')
        self.assertEqual(w.node.condition, '')

    def test__compare_value(self):
        """"""
        w = self.w
        w._compare_value('')
        self.assertEqual(w.node.value, '')

    def test__is_value(self):
        """"""
        w = self.w
        w._is_value('')
        self.assertEqual(w.node.value, '')

    def test__like_value(self):
        """"""
        w = self.w
        w._like_value('')
        self.assertEqual(w.node.value, '')

    def test__value_right_parenthesis(self):
        """"""
        w = self.w
        self.assertRaises(where.WhereError, w._value_right_parenthesis, *('',))
        w.node.up = True
        w._value_right_parenthesis('')

    def test__present_right_parenthesis(self):
        """"""
        w = self.w
        self.assertRaises(
            where.WhereError, w._present_right_parenthesis, *('',))
        w.node.up = True
        w._present_right_parenthesis('')

    def test__field_present(self):
        """"""
        w = self.w
        w._field_present('')

    def test__field_alphanumeric(self):
        """"""
        w = self.w
        w._field_alphanumeric('')
        self.assertEqual(w.node.alpha, False)
        self.assertEqual(w.node.num, False)
        w._field_alphanumeric('num')
        self.assertEqual(w.node.alpha, False)
        self.assertEqual(w.node.num, True)
        w._field_alphanumeric('alpha')
        self.assertEqual(w.node.alpha, True)
        self.assertEqual(w.node.num, False)

    def test__first_token_invert(self):
        """"""
        w = self.w
        w._first_token_invert('')
        self.assertEqual(w.node.not_phrase, True)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)
        self.assertIs(w.node.up, self.wc)
        self.assertIs(w.node, self.wc.down)

    def test__value_or_field_after_invert_01(self):
        """"""
        w = self.w
        w.node.condition = where.IS
        w._value_or_field_after_invert('')
        self.assertEqual(w.node.value, '')
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.condition, where.IS)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__value_or_field_after_invert_02(self):
        """"""
        w = self.w
        w._value_or_field_after_invert('')
        self.assertEqual(w.node.field, '')
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__value_or_field_after_invert_03(self):
        """"""
        w = self.w
        w._not = True
        w.node.condition = where.IS
        w._value_or_field_after_invert('')
        self.assertEqual(w.node.value, '')
        self.assertEqual(w.node.not_value, True)
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.condition, where.IS)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__value_or_field_after_invert_04(self):
        """"""
        w = self.w
        w._not = True
        w._value_or_field_after_invert('')
        self.assertEqual(w.node.field, '')
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.not_phrase, True)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__boolean_invert(self):
        """"""
        w = self.w
        w._boolean_invert('')
        self.assertEqual(w._not, True)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.not_condition, None)

    def test__field_invert(self):
        """"""
        w = self.w
        w._field_invert('')
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, True)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__invert_alphanumeric_01(self):
        """"""
        w = self.w
        w._invert_alphanumeric(where.ALPHA)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.num, False)
        self.assertEqual(w.node.alpha, True)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__invert_alphanumeric_02(self):
        """"""
        w = self.w
        w._invert_alphanumeric(where.NUM)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.num, True)
        self.assertEqual(w.node.alpha, False)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__boolean_alphanumeric_01(self):
        """"""
        w = self.w
        w._boolean_alphanumeric(where.ALPHA)
        self.assertEqual(w.node.field, w.node.left.field)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.not_phrase, w.node.left.not_phrase)
        self.assertEqual(w.node.operator, w.node.left.operator)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.num, False)
        self.assertEqual(w.node.alpha, True)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__boolean_alphanumeric_02(self):
        """"""
        w = self.w
        w._boolean_alphanumeric(where.NUM)
        self.assertEqual(w.node.field, w.node.left.field)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.not_phrase, w.node.left.not_phrase)
        self.assertEqual(w.node.operator, w.node.left.operator)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.num, True)
        self.assertEqual(w.node.alpha, False)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__invert_present(self):
        """"""
        w = self.w
        w._invert_present('')
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.condition, '')
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__boolean_present(self):
        """"""
        w = self.w
        w._boolean_present('')
        self.assertEqual(w.node.field, w.node.left.field)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.not_phrase, w.node.left.not_phrase)
        self.assertEqual(w.node.operator, w.node.left.operator)
        self.assertEqual(w.node.condition, '')
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, w.node.left.not_condition)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__boolean_is(self):
        """"""
        w = self.w
        w._boolean_is('')
        self.assertEqual(w.node.field, w.node.left.field)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.not_phrase, w.node.left.not_phrase)
        self.assertEqual(w.node.operator, w.node.left.operator)
        self.assertEqual(w.node.condition, '')
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__invert_like(self):
        """"""
        w = self.w
        w._invert_like('')
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.condition, '')
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__boolean_like(self):
        """"""
        w = self.w
        w._boolean_like('')
        self.assertEqual(w.node.field, w.node.left.field)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.not_phrase, w.node.left.not_phrase)
        self.assertEqual(w.node.operator, w.node.left.operator)
        self.assertEqual(w.node.condition, '')
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, w.node.left.not_condition)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__is_invert(self):
        """"""
        w = self.w
        w._is_invert('')
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.not_value, True)
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__value_or_field_after_boolean(self):
        """"""
        w = self.w
        w._value_or_field_after_boolean('')
        self.assertEqual(w._f_or_v, '')
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__first_token_left_parenthesis(self):
        """"""
        w = self.w
        w._first_token_left_parenthesis('')
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)
        self.assertIs(w.node.up.up, self.wc)
        self.assertIs(w.node.up, self.wc.down)
        self.assertIs(w.node, self.wc.down.down)

    def test__right_parenthesis_boolean(self):
        """"""
        w = self.w
        w._right_parenthesis_boolean('')
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, '')
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)
        self.assertIs(w.node, self.wc.right)
        self.assertIs(w.node.left, self.wc)
        self.assertIs(w.node.up, self.wc.up)

    def test__left_parenthesis_invert(self):
        """"""
        w = self.w
        w._left_parenthesis_invert('')
        self.assertEqual(w.node.not_phrase, True)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__left_parenthesis_left_parenthesis(self):
        """"""
        w = self.w
        w._left_parenthesis_left_parenthesis('')
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)
        self.assertIs(w.node.up, self.wc)
        self.assertIs(w.node, self.wc.down)
        self.assertIs(w.node.up.left, self.wc.left)

    def test__right_parenthesis_right_parenthesis(self):
        """"""
        w = self.w
        w.node.down = where.WhereClause()
        w.node.down.up = w.node
        w.node = w.node.down
        w._right_parenthesis_right_parenthesis('')
        for n in (w.node, w.node.down):
            self.assertEqual(n.not_phrase, None)
            self.assertEqual(n.operator, None)
            self.assertEqual(n.field, None)
            self.assertEqual(n.condition, None)
            self.assertEqual(n.value, None)
            self.assertEqual(n.not_condition, None)
            self.assertEqual(n.not_value, None)
            self.assertEqual(n.num, None)
            self.assertEqual(n.alpha, None)
            self.assertEqual(n.result, None)
            self.assertEqual(n.constraint, None)
        self.assertIs(w.node, self.wc)
        self.assertIs(w.node.down.up, self.wc)

    def test__finish_01(self):
        """"""
        w = self.w
        self.assertEqual(w._f_or_v, None)
        w._finish('')
        self.assertEqual(w._f_or_v, None)
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__finish_02(self):
        """"""
        w = self.w
        w._f_or_v = ''
        w._finish('')
        self.assertEqual(w._f_or_v, None)
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, '')
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__deferred_field_01(self):
        """"""
        w = self.w
        self.assertEqual(w._f_or_v, None)
        w._deferred_field()
        self.assertEqual(w._f_or_v, None)
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__deferred_field_02(self):
        """"""
        w = self.w
        w._f_or_v = ''
        w._deferred_field()
        self.assertEqual(w._f_or_v, None)
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, '')
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__deferred_not_phrase_01(self):
        """"""
        w = self.w
        self.assertEqual(w._not, None)
        w._deferred_not_phrase()
        self.assertEqual(w._not, None)
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__deferred_not_phrase_02(self):
        """"""
        w = self.w
        w._not = True
        w._deferred_not_phrase()
        self.assertEqual(w._not, None)
        self.assertEqual(w.node.not_phrase, True)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__deferred_not_condition_01(self):
        """"""
        w = self.w
        self.assertEqual(w._not, None)
        w._deferred_not_condition()
        self.assertEqual(w._not, None)
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__deferred_not_condition_02(self):
        """"""
        w = self.w
        w._not = True
        w._deferred_not_condition()
        self.assertEqual(w._not, None)
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, True)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__deferred_not_value_01(self):
        """"""
        w = self.w
        self.assertEqual(w._not, None)
        w._deferred_not_value()
        self.assertEqual(w._not, None)
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__deferred_not_value_02(self):
        """"""
        w = self.w
        w._not = True
        w._deferred_not_value()
        self.assertEqual(w._not, None)
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.not_value, True)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__deferred_value_01(self):
        """"""
        w = self.w
        self.assertEqual(w._f_or_v, None)
        w._deferred_value()
        self.assertEqual(w._f_or_v, None)
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__deferred_value_02(self):
        """"""
        w = self.w
        w._f_or_v = ''
        w._deferred_value()
        self.assertEqual(w._f_or_v, None)
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, '')
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__copy_pre_field(self):
        """"""
        w = self.w
        w.node.left.not_phrase = ''
        w._copy_pre_field()
        self.assertEqual(w.node.not_phrase, '')
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__copy_pre_compare(self):
        """"""
        w = self.w
        w.node.left.num = True
        w.node.left.alpha = False
        w._copy_pre_compare()
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.num, True)
        self.assertEqual(w.node.alpha, False)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__copy_pre_range(self):
        """"""
        w = self.w
        w.node.left.num = False
        w.node.left.alpha = True
        w._copy_pre_range()
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.num, False)
        self.assertEqual(w.node.alpha, True)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__copy_pre_alphanumeric(self):
        """"""
        w = self.w
        w.node.left.not_condition = ''
        w._copy_pre_alphanumeric()
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, '')
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__copy_pre_present(self):
        """"""
        w = self.w
        w.node.left.not_condition = 'x'
        w._copy_pre_present()
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, 'x')
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__copy_pre_is(self):
        """"""
        w = self.w
        w.node.left.field = ''
        w._copy_pre_is()
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, '')
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__copy_pre_like(self):
        """"""
        w = self.w
        w.node.left.not_condition = ''
        w._copy_pre_like()
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, None)
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, '')
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__copy_pre_not_condition(self):
        """"""
        w = self.w
        w.node.left.field = 'ff'
        w._copy_pre_not_condition()
        self.assertEqual(w.node.not_phrase, None)
        self.assertEqual(w.node.operator, None)
        self.assertEqual(w.node.field, 'ff')
        self.assertEqual(w.node.condition, None)
        self.assertEqual(w.node.value, None)
        self.assertEqual(w.node.not_condition, None)
        self.assertEqual(w.node.not_value, None)
        self.assertEqual(w.node.num, None)
        self.assertEqual(w.node.alpha, None)
        self.assertEqual(w.node.result, None)
        self.assertEqual(w.node.constraint, None)

    def test__copy_pre_value_01(self):
        """"""
        w = self.w
        w.node.left.condition = where.PRESENT
        self.assertRaises(where.WhereError, w._copy_pre_value)

    def test__copy_pre_value_02(self):
        """"""
        w = self.w
        w.node.left.condition = (where.FROM, where.TO)
        self.assertRaises(where.WhereError, w._copy_pre_value)

    def test__copy_pre_value_03(self):
        """"""
        w = self.w
        w.node.left.condition = (where.FROM, where.BELOW)
        self.assertRaises(where.WhereError, w._copy_pre_value)

    def test__copy_pre_value_04(self):
        """"""
        w = self.w
        w.node.left.condition = (where.ABOVE, where.TO)
        self.assertRaises(where.WhereError, w._copy_pre_value)

    def test__copy_pre_value_05(self):
        """"""
        w = self.w
        w.node.left.condition = (where.ABOVE, where.BELOW)
        self.assertRaises(where.WhereError, w._copy_pre_value)

    def test__copy_pre_value_06(self):
        """"""
        w = self.w
        w.node.left.condition = where.LIKE
        w._copy_pre_value()
        self.assertEqual(w.node.condition, where.LIKE)

    def test__copy_pre_value_07(self):
        """"""
        w = self.w
        w.node.left.condition = where.IS
        w._copy_pre_value()
        self.assertEqual(w.node.condition, where.IS)

    def test__copy_pre_value_08(self):
        """"""
        w = self.w
        w.node.left.condition = where.EQ
        w._copy_pre_value()
        self.assertEqual(w.node.condition, where.EQ)


class Where_lex_phraseTC(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test__no_keywords(self):
        w = where.Where('this statement has no keywords')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['this statement has no keywords', None])

    def test__no_keywords_mixed(self):
        w = where.Where('this Statement has no keYWOrds')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['this Statement has no keYWOrds', None])

    def test_is_upper(self):
        w = where.Where('fieldname IS value')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['fieldname', 'is', 'value', None])

    def test_non_keywords_mixed(self):
        w = where.Where('fieLdnAme is VALUE')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['fieLdnAme', 'is', 'VALUE', None])

    def test_like_mixed(self):
        w = where.Where('fieldname lIKe pattern')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['fieldname', 'like', 'pattern', None])

    def test_parentheses(self):
        w = where.Where('(some text)')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['(', 'some text', ')', None])

    def test_is(self):
        w = where.Where('fieldname is value')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['fieldname', 'is', 'value', None])

    def test_like(self):
        w = where.Where('fieldname like pattern')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['fieldname', 'like', 'pattern', None])

    def test_present(self):
        w = where.Where('fieldname present')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['fieldname', 'present', None])

    def test_alpha_eq(self):
        w = where.Where('floor alpha eq basement')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'alpha', 'eq', 'basement', None])

    def test_alpha_ne(self):
        w = where.Where('floor alpha ne first')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'alpha', 'ne', 'first', None])

    def test_alpha_gt(self):
        w = where.Where('floor alpha gt ground')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'alpha', 'gt', 'ground', None])

    def test_alpha_lt(self):
        w = where.Where('houses in street alpha lt thirty three')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['houses in street', 'alpha', 'lt', 'thirty three', None])

    def test_alpha_le(self):
        w = where.Where('floor alpha le third')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'alpha', 'le', 'third', None])

    def test_alpha_ge(self):
        w = where.Where('floor alpha ge second')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'alpha', 'ge', 'second', None])

    def test_alpha_before(self):
        w = where.Where('house room alpha before attic')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'alpha', 'before', 'attic', None])

    def test_alpha_after(self):
        w = where.Where('house room alpha after the ground floor')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'alpha', 'after', 'the ground floor', None])

    def test_alpha_from_to(self):
        w = where.Where('house room alpha from the floor to the roof')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'alpha', 'from', 'the floor', 'to',
             'the roof', None])

    def test_alpha_from_below(self):
        w = where.Where('house room alpha from the floor below the roof')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'alpha', 'from', 'the floor', 'below',
             'the roof', None])

    def test_alpha_above_to(self):
        w = where.Where('house room alpha above the floor to the roof')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'alpha', 'above', 'the floor', 'to',
             'the roof', None])

    def test_alpha_above_below(self):
        w = where.Where('house room alpha above the floor below the roof')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'alpha', 'above', 'the floor', 'below',
             'the roof', None])

    def test_num_eq(self):
        w = where.Where('floor num eq basement')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'num', 'eq', 'basement', None])

    def test_num_ne(self):
        w = where.Where('floor num ne first')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'num', 'ne', 'first', None])

    def test_num_gt(self):
        w = where.Where('floor num gt ground')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'num', 'gt', 'ground', None])

    def test_num_lt(self):
        w = where.Where('houses in street num lt thirty three')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['houses in street', 'num', 'lt', 'thirty three', None])

    def test_num_le(self):
        w = where.Where('floor num le third')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'num', 'le', 'third', None])

    def test_num_ge(self):
        w = where.Where('floor num ge second')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'num', 'ge', 'second', None])

    def test_num_before(self):
        w = where.Where('house room num before attic')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'num', 'before', 'attic', None])

    def test_num_after(self):
        w = where.Where('house room num after the ground floor')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'num', 'after', 'the ground floor', None])

    def test_num_from_to(self):
        w = where.Where('house room num from the floor to the roof')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'num', 'from', 'the floor', 'to',
             'the roof', None])

    def test_num_from_below(self):
        w = where.Where('house room num from the floor below the roof')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'num', 'from', 'the floor', 'below',
             'the roof', None])

    def test_num_above_to(self):
        w = where.Where('house room num above the floor to the roof')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'num', 'above', 'the floor', 'to', 'the roof', None])

    def test_num_above_below(self):
        w = where.Where('house room num above the floor below the roof')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'num', 'above', 'the floor', 'below',
             'the roof', None])

    def test_eq(self):
        w = where.Where('floor eq basement')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'eq', 'basement', None])

    def test_ne(self):
        w = where.Where('floor ne first')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'ne', 'first', None])

    def test_gt(self):
        w = where.Where('floor gt ground')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'gt', 'ground', None])

    def test_lt(self):
        w = where.Where('houses in street lt thirty three')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['houses in street', 'lt', 'thirty three', None])

    def test_le(self):
        w = where.Where('floor le third')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'le', 'third', None])

    def test_ge(self):
        w = where.Where('floor ge second')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'ge', 'second', None])

    def test_before(self):
        w = where.Where('house room before attic')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'before', 'attic', None])

    def test_after(self):
        w = where.Where('house room after the ground floor')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'after', 'the ground floor', None])

    def test_from_to(self):
        w = where.Where('house room from the floor to the roof')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'from', 'the floor', 'to', 'the roof', None])

    def test_from_below(self):
        w = where.Where('house room from the floor below the roof')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'from', 'the floor', 'below', 'the roof', None])

    def test_above_to(self):
        w = where.Where('house room above the floor to the roof')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'above', 'the floor', 'to', 'the roof', None])

    def test_above_below(self):
        w = where.Where('house room above the floor below the roof')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'above', 'the floor', 'below', 'the roof', None])

    def test_is_not(self):
        w = where.Where('fieldname is not value')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['fieldname', 'is', 'not', 'value', None])

    def test_not_like(self):
        w = where.Where('fieldname not like pattern')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['fieldname', 'not', 'like', 'pattern', None])

    def test_not_present(self):
        w = where.Where('fieldname not present')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['fieldname', 'not', 'present', None])

    def test_not_alpha_eq(self):
        w = where.Where('floor not alpha eq basement')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'not', 'alpha', 'eq', 'basement', None])

    def test_not_alpha_ne(self):
        w = where.Where('floor not alpha ne first')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'not', 'alpha', 'ne', 'first', None])

    def test_not_alpha_gt(self):
        w = where.Where('floor not alpha gt ground')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'not', 'alpha', 'gt', 'ground', None])

    def test_not_alpha_lt(self):
        w = where.Where('houses in street not alpha lt thirty three')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['houses in street', 'not', 'alpha', 'lt', 'thirty three', None])

    def test_not_alpha_le(self):
        w = where.Where('floor not alpha le third')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'not', 'alpha', 'le', 'third', None])

    def test_not_alpha_ge(self):
        w = where.Where('floor not alpha ge second')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'not', 'alpha', 'ge', 'second', None])

    def test_not_alpha_before(self):
        w = where.Where('house room not alpha before attic')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'not', 'alpha', 'before', 'attic', None])

    def test_not_alpha_after(self):
        w = where.Where('house room not alpha after the ground floor')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'not', 'alpha', 'after', 'the ground floor', None])

    def test_not_alpha_from_to(self):
        w = where.Where('house room not alpha from the floor to the roof')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'not', 'alpha',
             'from', 'the floor', 'to', 'the roof', None])

    def test_not_alpha_from_below(self):
        w = where.Where('house room not alpha from the floor below the roof')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'not', 'alpha',
             'from', 'the floor', 'below', 'the roof', None])

    def test_not_alpha_above_to(self):
        w = where.Where('house room not alpha above the floor to the roof')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'not', 'alpha',
             'above', 'the floor', 'to', 'the roof', None])

    def test_not_alpha_above_below(self):
        w = where.Where('house room not alpha above the floor below the roof')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'not', 'alpha',
             'above', 'the floor', 'below', 'the roof', None])

    def test_not_num_eq(self):
        w = where.Where('floor not num eq basement')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'not', 'num', 'eq', 'basement', None])

    def test_not_num_ne(self):
        w = where.Where('floor not num ne first')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'not', 'num', 'ne', 'first', None])

    def test_not_num_gt(self):
        w = where.Where('floor not num gt ground')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'not', 'num', 'gt', 'ground', None])

    def test_not_num_lt(self):
        w = where.Where('houses in street not num lt thirty three')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['houses in street', 'not', 'num', 'lt', 'thirty three', None])

    def test_not_num_le(self):
        w = where.Where('floor not num le third')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'not', 'num', 'le', 'third', None])

    def test_not_num_ge(self):
        w = where.Where('floor not num ge second')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'not', 'num', 'ge', 'second', None])

    def test_not_num_before(self):
        w = where.Where('house room not num before attic')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'not', 'num', 'before', 'attic', None])

    def test_not_num_after(self):
        w = where.Where('house room not num after the ground floor')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'not', 'num', 'after', 'the ground floor', None])

    def test_not_num_from_to(self):
        w = where.Where('house room not num from the floor to the roof')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'not', 'num',
             'from', 'the floor', 'to', 'the roof', None])

    def test_not_num_from_below(self):
        w = where.Where('house room not num from the floor below the roof')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'not', 'num',
             'from', 'the floor', 'below', 'the roof', None])

    def test_not_num_above_to(self):
        w = where.Where('house room not num above the floor to the roof')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'not', 'num',
             'above', 'the floor', 'to', 'the roof', None])

    def test_not_num_above_below(self):
        w = where.Where('house room not num above the floor below the roof')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'not', 'num',
             'above', 'the floor', 'below', 'the roof', None])

    def test_not_eq(self):
        w = where.Where('floor not eq basement')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'not', 'eq', 'basement', None])

    def test_not_ne(self):
        w = where.Where('floor not ne first')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'not', 'ne', 'first', None])

    def test_not_gt(self):
        w = where.Where('floor not gt ground')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'not', 'gt', 'ground', None])

    def test_not_lt(self):
        w = where.Where('houses in street not lt thirty three')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['houses in street', 'not', 'lt', 'thirty three', None])

    def test_not_le(self):
        w = where.Where('floor not le third')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'not', 'le', 'third', None])

    def test_not_ge(self):
        w = where.Where('floor not ge second')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['floor', 'not', 'ge', 'second', None])

    def test_not_before(self):
        w = where.Where('house room not before attic')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'not', 'before', 'attic', None])

    def test_not_after(self):
        w = where.Where('house room not after the ground floor')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'not', 'after', 'the ground floor', None])

    def test_not_from_to(self):
        w = where.Where('house room not from the floor to the roof')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'not', 'from', 'the floor', 'to', 'the roof', None])

    def test_not_from_below(self):
        w = where.Where('house room not from the floor below the roof')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'not', 'from', 'the floor', 'below',
             'the roof', None])

    def test_not_above_to(self):
        w = where.Where('house room not above the floor to the roof')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'not', 'above', 'the floor', 'to', 'the roof', None])

    def test_not_above_below(self):
        w = where.Where('house room not above the floor below the roof')
        w.lex()
        self.assertEqual(
            w.tokens,
            ['house room', 'not', 'above', 'the floor', 'below',
             'the roof', None])


class Where_parse_phraseTC(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test__no_keywords(self):
        self._cwc(
            where.Where('this statement has no keywords'),
            field='this statement has no keywords')

    def test__no_keywords_mixed(self):
        self._cwc(
            where.Where('this Statement has no keYWOrds'),
            field='this Statement has no keYWOrds')

    def test_is_upper(self):
        self._cwc(
            where.Where('fieldname IS value'),
            field='fieldname',
            condition='is',
            value='value')

    def test_non_keywords_mixed(self):
        self._cwc(
            where.Where('fieLdnAme is VALUE'),
            field='fieLdnAme',
            condition='is',
            value='VALUE')

    def test_like_mixed(self):
        self._cwc(
            where.Where('fieldname lIKe pattern'),
            field='fieldname',
            condition='like',
            value='pattern')

    def test_is(self):
        self._cwc(
            where.Where('fieldname is value'),
            field='fieldname',
            condition='is',
            value='value')

    def test_like(self):
        self._cwc(
            where.Where('fieldname like pattern'),
            field='fieldname',
            condition='like',
            value='pattern')

    def test_present(self):
        self._cwc(
            where.Where('fieldname present'),
            field='fieldname',
            condition='present')

    def test_alpha_eq(self):
        self._cwc(
            where.Where('floor alpha eq basement'),
            field='floor',
            condition='eq',
            value='basement',
            num=False,
            alpha=True)

    def test_alpha_ne(self):
        self._cwc(
            where.Where('floor alpha ne first'),
            field='floor',
            condition='ne',
            value='first',
            num=False,
            alpha=True)

    def test_alpha_gt(self):
        self._cwc(
            where.Where('floor alpha gt ground'),
            field='floor',
            condition='gt',
            value='ground',
            num=False,
            alpha=True)

    def test_alpha_lt(self):
        self._cwc(
            where.Where('houses in street alpha lt thirty three'),
            field='houses in street',
            condition='lt',
            value='thirty three',
            num=False,
            alpha=True)

    def test_alpha_le(self):
        self._cwc(
            where.Where('floor alpha le third'),
            field='floor',
            condition='le',
            value='third',
            num=False,
            alpha=True)

    def test_alpha_ge(self):
        self._cwc(
            where.Where('floor alpha ge second'),
            field='floor',
            condition='ge',
            value='second',
            num=False,
            alpha=True)

    def test_alpha_before(self):
        self._cwc(
            where.Where('house room alpha before attic'),
            field='house room',
            condition='before',
            value='attic',
            num=False,
            alpha=True)

    def test_alpha_after(self):
        self._cwc(
            where.Where('house room alpha after the ground floor'),
            field='house room',
            condition='after',
            value='the ground floor',
            num=False,
            alpha=True)

    def test_alpha_from_to(self):
        self._cwc(
            where.Where('house room alpha from the floor to the roof'),
            field='house room',
            condition=('from', 'to'),
            value=('the floor', 'the roof'),
            num=False,
            alpha=True)

    def test_alpha_from_below(self):
        self._cwc(
            where.Where('house room alpha from the floor below the roof'),
            field='house room',
            condition=('from', 'below'),
            value=('the floor', 'the roof'),
            num=False,
            alpha=True)

    def test_alpha_above_to(self):
        self._cwc(
            where.Where('house room alpha above the floor to the roof'),
            field='house room',
            condition=('above', 'to'),
            value=('the floor', 'the roof'),
            num=False,
            alpha=True)

    def test_alpha_above_below(self):
        self._cwc(
            where.Where('house room alpha above the floor below the roof'),
            field='house room',
            condition=('above', 'below'),
            value=('the floor', 'the roof'),
            num=False,
            alpha=True)

    def test_num_eq(self):
        self._cwc(
            where.Where('floor num eq basement'),
            field='floor',
            condition='eq',
            value='basement',
            num=True,
            alpha=False)

    def test_num_ne(self):
        self._cwc(
            where.Where('floor num ne first'),
            field='floor',
            condition='ne',
            value='first',
            num=True,
            alpha=False)

    def test_num_gt(self):
        self._cwc(
            where.Where('floor num gt ground'),
            field='floor',
            condition='gt',
            value='ground',
            num=True,
            alpha=False)

    def test_num_lt(self):
        self._cwc(
            where.Where('houses in street num lt thirty three'),
            field='houses in street',
            condition='lt',
            value='thirty three',
            num=True,
            alpha=False)

    def test_num_le(self):
        self._cwc(
            where.Where('floor num le third'),
            field='floor',
            condition='le',
            value='third',
            num=True,
            alpha=False)

    def test_num_ge(self):
        self._cwc(
            where.Where('floor num ge second'),
            field='floor',
            condition='ge',
            value='second',
            num=True,
            alpha=False)

    def test_num_before(self):
        self._cwc(
            where.Where('house room num before attic'),
            field='house room',
            condition='before',
            value='attic',
            num=True,
            alpha=False)

    def test_num_after(self):
        self._cwc(
            where.Where('house room num after the ground floor'),
            field='house room',
            condition='after',
            value='the ground floor',
            num=True,
            alpha=False)

    def test_num_from_to(self):
        self._cwc(
            where.Where('house room num from the floor to the roof'),
            field='house room',
            condition=('from', 'to'),
            value=('the floor', 'the roof'),
            num=True,
            alpha=False)

    def test_num_from_below(self):
        self._cwc(
            where.Where('house room num from the floor below the roof'),
            field='house room',
            condition=('from', 'below'),
            value=('the floor', 'the roof'),
            num=True,
            alpha=False)

    def test_num_above_to(self):
        self._cwc(
            where.Where('house room num above the floor to the roof'),
            field='house room',
            condition=('above', 'to'),
            value=('the floor', 'the roof'),
            num=True,
            alpha=False)

    def test_num_above_below(self):
        self._cwc(
            where.Where('house room num above the floor below the roof'),
            field='house room',
            condition=('above', 'below'),
            value=('the floor', 'the roof'),
            num=True,
            alpha=False)

    def test_eq(self):
        self._cwc(
            where.Where('floor eq basement'),
            field='floor',
            condition='eq',
            value='basement')

    def test_ne(self):
        self._cwc(
            where.Where('floor ne first'),
            field='floor',
            condition='ne',
            value='first')

    def test_gt(self):
        self._cwc(
            where.Where('floor gt ground'),
            field='floor',
            condition='gt',
            value='ground')

    def test_lt(self):
        self._cwc(
            where.Where('houses in street lt thirty three'),
            field='houses in street',
            condition='lt',
            value='thirty three')

    def test_le(self):
        self._cwc(
            where.Where('floor le third'),
            field='floor',
            condition='le',
            value='third')

    def test_ge(self):
        self._cwc(
            where.Where('floor ge second'),
            field='floor',
            condition='ge',
            value='second')

    def test_before(self):
        self._cwc(
            where.Where('house room before attic'),
            field='house room',
            condition='before',
            value='attic')

    def test_after(self):
        self._cwc(
            where.Where('house room after the ground floor'),
            field='house room',
            condition='after',
            value='the ground floor')

    def test_from_to(self):
        self._cwc(
            where.Where('house room from the floor to the roof'),
            field='house room',
            condition=('from', 'to'),
            value=('the floor', 'the roof'))

    def test_from_below(self):
        self._cwc(
            where.Where('house room from the floor below the roof'),
            field='house room',
            condition=('from', 'below'),
            value=('the floor', 'the roof'))

    def test_above_to(self):
        self._cwc(
            where.Where('house room above the floor to the roof'),
            field='house room',
            condition=('above', 'to'),
            value=('the floor', 'the roof'))

    def test_above_below(self):
        self._cwc(
            where.Where('house room above the floor below the roof'),
            field='house room',
            condition=('above', 'below'),
            value=('the floor', 'the roof'))

    def test_is_not(self):
        self._cwc(
            where.Where('fieldname is not value'),
            field='fieldname',
            condition='is',
            value='value',
            not_value=True)

    def test_not_like(self):
        self._cwc(
            where.Where('fieldname not like pattern'),
            field='fieldname',
            condition='like',
            value='pattern',
            not_condition=True)

    def test_not_present(self):
        self._cwc(
            where.Where('fieldname not present'),
            field='fieldname',
            condition='present',
            not_condition=True)

    def test_not_alpha_eq(self):
        self._cwc(
            where.Where('floor not alpha eq basement'),
            field='floor',
            condition='eq',
            value='basement',
            not_condition=True,
            num=False,
            alpha=True)

    def test_not_alpha_ne(self):
        self._cwc(
            where.Where('floor not alpha ne first'),
            field='floor',
            condition='ne',
            value='first',
            not_condition=True,
            num=False,
            alpha=True)

    def test_not_alpha_gt(self):
        self._cwc(
            where.Where('floor not alpha gt ground'),
            field='floor',
            condition='gt',
            value='ground',
            not_condition=True,
            num=False,
            alpha=True)

    def test_not_alpha_lt(self):
        self._cwc(
            where.Where('houses in street not alpha lt thirty three'),
            field='houses in street',
            condition='lt',
            value='thirty three',
            not_condition=True,
            num=False,
            alpha=True)

    def test_not_alpha_le(self):
        self._cwc(
            where.Where('floor not alpha le third'),
            field='floor',
            condition='le',
            value='third',
            not_condition=True,
            num=False,
            alpha=True)

    def test_not_alpha_ge(self):
        self._cwc(
            where.Where('floor not alpha ge second'),
            field='floor',
            condition='ge',
            value='second',
            not_condition=True,
            num=False,
            alpha=True)

    def test_not_alpha_before(self):
        self._cwc(
            where.Where('house room not alpha before attic'),
            field='house room',
            condition='before',
            value='attic',
            not_condition=True,
            num=False,
            alpha=True)

    def test_not_alpha_after(self):
        self._cwc(
            where.Where('house room not alpha after the ground floor'),
            field='house room',
            condition='after',
            value='the ground floor',
            not_condition=True,
            num=False,
            alpha=True)

    def test_not_alpha_from_to(self):
        self._cwc(
            where.Where('house room not alpha from the floor to the roof'),
            field='house room',
            condition=('from', 'to'),
            value=('the floor', 'the roof'),
            not_condition=True,
            num=False,
            alpha=True)

    def test_not_alpha_from_below(self):
        self._cwc(
            where.Where('house room not alpha from the floor below the roof'),
            field='house room',
            condition=('from', 'below'),
            value=('the floor', 'the roof'),
            not_condition=True,
            num=False,
            alpha=True)

    def test_not_alpha_above_to(self):
        self._cwc(
            where.Where('house room not alpha above the floor to the roof'),
            field='house room',
            condition=('above', 'to'),
            value=('the floor', 'the roof'),
            not_condition=True,
            num=False,
            alpha=True)

    def test_not_alpha_above_below(self):
        self._cwc(
            where.Where('house room not alpha above the floor below the roof'),
            field='house room',
            condition=('above', 'below'),
            value=('the floor', 'the roof'),
            not_condition=True,
            num=False,
            alpha=True)

    def test_not_num_eq(self):
        self._cwc(
            where.Where('floor not num eq basement'),
            field='floor',
            condition='eq',
            value='basement',
            not_condition=True,
            num=True,
            alpha=False)

    def test_not_num_ne(self):
        self._cwc(
            where.Where('floor not num ne first'),
            field='floor',
            condition='ne',
            value='first',
            not_condition=True,
            num=True,
            alpha=False)

    def test_not_num_gt(self):
        self._cwc(
            where.Where('floor not num gt ground'),
            field='floor',
            condition='gt',
            value='ground',
            not_condition=True,
            num=True,
            alpha=False)

    def test_not_num_lt(self):
        self._cwc(
            where.Where('houses in street not num lt thirty three'),
            field='houses in street',
            condition='lt',
            value='thirty three',
            not_condition=True,
            num=True,
            alpha=False)

    def test_not_num_le(self):
        self._cwc(
            where.Where('floor not num le third'),
            field='floor',
            condition='le',
            value='third',
            not_condition=True,
            num=True,
            alpha=False)

    def test_not_num_ge(self):
        self._cwc(
            where.Where('floor not num ge second'),
            field='floor',
            condition='ge',
            value='second',
            not_condition=True,
            num=True,
            alpha=False)

    def test_not_num_before(self):
        self._cwc(
            where.Where('house room not num before attic'),
            field='house room',
            condition='before',
            value='attic',
            not_condition=True,
            num=True,
            alpha=False)

    def test_not_num_after(self):
        self._cwc(
            where.Where('house room not num after the ground floor'),
            field='house room',
            condition='after',
            value='the ground floor',
            not_condition=True,
            num=True,
            alpha=False)

    def test_not_num_from_to(self):
        self._cwc(
            where.Where('house room not num from the floor to the roof'),
            field='house room',
            condition=('from', 'to'),
            value=('the floor', 'the roof'),
            not_condition=True,
            num=True,
            alpha=False)

    def test_not_num_from_below(self):
        self._cwc(
            where.Where('house room not num from the floor below the roof'),
            field='house room',
            condition=('from', 'below'),
            value=('the floor', 'the roof'),
            not_condition=True,
            num=True,
            alpha=False)

    def test_not_num_above_to(self):
        self._cwc(
            where.Where('house room not num above the floor to the roof'),
            field='house room',
            condition=('above', 'to'),
            value=('the floor', 'the roof'),
            not_condition=True,
            num=True,
            alpha=False)

    def test_not_num_above_below(self):
        self._cwc(
            where.Where('house room not num above the floor below the roof'),
            field='house room',
            condition=('above', 'below'),
            value=('the floor', 'the roof'),
            not_condition=True,
            num=True,
            alpha=False)

    def test_not_eq(self):
        self._cwc(
            where.Where('floor not eq basement'),
            field='floor',
            condition='eq',
            value='basement',
            not_condition=True)

    def test_not_ne(self):
        self._cwc(
            where.Where('floor not ne first'),
            field='floor',
            condition='ne',
            value='first',
            not_condition=True)

    def test_not_gt(self):
        self._cwc(
            where.Where('floor not gt ground'),
            field='floor',
            condition='gt',
            value='ground',
            not_condition=True)

    def test_not_lt(self):
        self._cwc(
            where.Where('houses in street not lt thirty three'),
            field='houses in street',
            condition='lt',
            value='thirty three',
            not_condition=True)

    def test_not_le(self):
        self._cwc(
            where.Where('floor not le third'),
            field='floor',
            condition='le',
            value='third',
            not_condition=True)

    def test_not_ge(self):
        self._cwc(
            where.Where('floor not ge second'),
            field='floor',
            condition='ge',
            value='second',
            not_condition=True)

    def test_not_before(self):
        self._cwc(
            where.Where('house room not before attic'),
            field='house room',
            condition='before',
            value='attic',
            not_condition=True)

    def test_not_after(self):
        self._cwc(
            where.Where('house room not after the ground floor'),
            field='house room',
            condition='after',
            value='the ground floor',
            not_condition=True)

    def test_not_from_to(self):
        self._cwc(
            where.Where('house room not from the floor to the roof'),
            field='house room',
            condition=('from', 'to'),
            value=('the floor', 'the roof'),
            not_condition=True)

    def test_not_from_below(self):
        self._cwc(
            where.Where('house room not from the floor below the roof'),
            field='house room',
            condition=('from', 'below'),
            value=('the floor', 'the roof'),
            not_condition=True)

    def test_not_above_to(self):
        self._cwc(
            where.Where('house room not above the floor to the roof'),
            field='house room',
            condition=('above', 'to'),
            value=('the floor', 'the roof'),
            not_condition=True)

    def test_not_above_below(self):
        self._cwc(
            where.Where('house room not above the floor below the roof'),
            field='house room',
            condition=('above', 'below'),
            value=('the floor', 'the roof'),
            not_condition=True)

    def _cwc(
        self,
        where_clause,
        left=None,
        right=None,
        up=None,
        down=None,
        operator=None,
        field=None,
        condition=None,
        value=None,
        not_phrase=None,
        not_condition=None,
        not_value=None,
        num=None,
        alpha=None):
        where_clause.lex()
        where_clause.parse()
        ae = self.assertEqual
        wc = where_clause.node.down
        ae(wc.left, left)
        ae(wc.right, right)
        ae(wc.up, wc.get_root())
        ae(wc.down, down)
        ae(wc.operator, operator)
        ae(wc.field, field)
        ae(wc.condition, condition)
        ae(wc.value, value)
        ae(wc.not_phrase, not_phrase)
        ae(wc.not_condition, not_condition)
        ae(wc.not_value, not_value)
        ae(wc.num, num)
        ae(wc.alpha, alpha)


class Where_parse_multi_phraseTC(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_two_phrases(self):
        self._cwc(
            where.Where('f is a or f is b'),
            [dict(),
             dict(field='f', condition='is', value='a'),
             dict(field='f', condition='is', value='b', operator='or'),
             ])

    def test_all_five_simple_positive_phrases(self):
        self._cwc(
            where.Where(
                ' '.join(('f is a or f like b or f present or',
                          'f eq c or f from d to e or f is f',
                          ))),
            [dict(),
             dict(field='f', condition='is', value='a'),
             dict(field='f', condition='like', value='b', operator='or'),
             dict(field='f', condition='present', operator='or'),
             dict(field='f', condition='eq', value='c', operator='or'),
             dict(field='f',
                  condition=('from', 'to'), value=('d', 'e'), operator='or'),
             dict(field='f', condition='is', value='f', operator='or'),
             ])

    def test_all_five_simple_negative_phrases(self):
        self._cwc(
            where.Where(
                ' '.join(('f is not a or f not like b or f not present or',
                          'f not eq c or f not from d to e or f is not f',
                          ))),
            [dict(),
             dict(field='f', condition='is', value='a', not_value=True),
             dict(field='f', condition='like', value='b',
                  operator='or', not_condition=True),
             dict(field='f', condition='present',
                  operator='or', not_condition=True),
             dict(field='f', condition='eq', value='c',
                  operator='or', not_condition=True),
             dict(field='f',
                  condition=('from', 'to'), value=('d', 'e'),
                  operator='or', not_condition=True),
             dict(field='f', condition='is', value='f',
                  operator='or', not_value=True),
             ])

    def test_all_five_simple_positive_num_phrases(self):
        self._cwc(
            where.Where(
                ' '.join(('f is a or f like b or f present or',
                          'f num eq c or f num from d to e or f is f',
                          ))),
            [dict(),
             dict(field='f', condition='is', value='a'),
             dict(field='f', condition='like', value='b', operator='or'),
             dict(field='f', condition='present', operator='or'),
             dict(field='f', condition='eq', value='c', operator='or',
                  num=True, alpha=False),
             dict(field='f',
                  condition=('from', 'to'), value=('d', 'e'),
                  operator='or', num=True, alpha=False),
             dict(field='f', condition='is', value='f', operator='or'),
             ])

    def test_all_five_simple_negative_num_phrases(self):
        self._cwc(
            where.Where(
                ' '.join(('f is not a or f not like b or f not present or',
                          'f not num eq c or f not num from d to e or',
                          'f is not f',
                          ))),
            [dict(),
             dict(field='f', condition='is', value='a', not_value=True),
             dict(field='f', condition='like', value='b',
                  operator='or', not_condition=True),
             dict(field='f', condition='present',
                  operator='or', not_condition=True),
             dict(field='f', condition='eq', value='c',
                  operator='or', not_condition=True, num=True, alpha=False),
             dict(field='f',
                  condition=('from', 'to'), value=('d', 'e'),
                  operator='or', not_condition=True, num=True, alpha=False),
             dict(field='f', condition='is', value='f',
                  operator='or', not_value=True),
             ])

    def test_all_five_simple_positive_alpha_phrases(self):
        self._cwc(
            where.Where(
                ' '.join(('f is a or f like b or f present or',
                          'f alpha eq c or f alpha from d to e or f is f',
                          ))),
            [dict(),
             dict(field='f', condition='is', value='a'),
             dict(field='f', condition='like', value='b', operator='or'),
             dict(field='f', condition='present', operator='or'),
             dict(field='f', condition='eq', value='c', operator='or',
                  num=False, alpha=True),
             dict(field='f',
                  condition=('from', 'to'), value=('d', 'e'),
                  operator='or', num=False, alpha=True),
             dict(field='f', condition='is', value='f', operator='or'),
             ])

    def test_all_five_simple_negative_alpha_phrases(self):
        self._cwc(
            where.Where(
                ' '.join(('f is not a or f not like b or f not present or',
                          'f not alpha eq c or f not alpha from d to e or',
                          'f is not f',
                          ))),
            [dict(),
             dict(field='f', condition='is', value='a', not_value=True),
             dict(field='f', condition='like', value='b',
                  operator='or', not_condition=True),
             dict(field='f', condition='present',
                  operator='or', not_condition=True),
             dict(field='f', condition='eq', value='c',
                  operator='or', not_condition=True, num=False, alpha=True),
             dict(field='f',
                  condition=('from', 'to'), value=('d', 'e'),
                  operator='or', not_condition=True, num=False, alpha=True),
             dict(field='f', condition='is', value='f',
                  operator='or', not_value=True),
             ])

    def test_all_three_simple_boolean_phrases(self):
        self._cwc(
            where.Where(
                ' '.join(('f is a or f like b or f present and',
                          'f eq c and f from d to e nor f is f nor f le g',
                          ))),
            [dict(),
             dict(field='f', condition='is', value='a'),
             dict(field='f', condition='like', value='b', operator='or'),
             dict(field='f', condition='present', operator='or'),
             dict(field='f', condition='eq', value='c', operator='and'),
             dict(field='f',
                  condition=('from', 'to'), value=('d', 'e'), operator='and'),
             dict(field='f', condition='is', value='f', operator='nor'),
             dict(field='f', condition='le', value='g', operator='nor'),
             ])

    def _cwc(
        self,
        where_clause,
        expected_clauses):
        where_clause.lex()
        where_clause.parse()
        ae = self.assertEqual
        for e, c in enumerate(
            where_clause.node.get_clauses_from_root_in_walk_order()):
            wc = where.WhereClause()
            wc.__dict__.update(expected_clauses[e])
            ae(wc.operator, c.operator)
            ae(wc.field, c.field)
            ae(wc.condition, c.condition)
            ae(wc.value, c.value)
            ae(wc.not_phrase, c.not_phrase)
            ae(wc.not_condition, c.not_condition)
            ae(wc.not_value, c.not_value)
            ae(wc.num, c.num)
            ae(wc.alpha, c.alpha)


class Where_parse_not_phrasesTC(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_not_phrase_is(self):
        self._cwc(
            where.Where('not f is a'),
            [dict(),
             dict(field='f', condition='is', value='a', not_phrase=True),
             ])

    def test_not_phrase_is_not(self):
        self._cwc(
            where.Where('not f is not a'),
            [dict(),
             dict(field='f', condition='is', value='a',
                  not_phrase=True, not_value=True),
             ])

    def test_not_phrase_like(self):
        self._cwc(
            where.Where('not f like b'),
            [dict(),
             dict(field='f', condition='like', value='b',
                  not_phrase=True),
             ])

    def test_not_phrase_not_like(self):
        self._cwc(
            where.Where('not f not like b'),
            [dict(),
             dict(field='f', condition='like', value='b',
                  not_phrase=True, not_condition=True),
             ])

    def test_not_phrase_present(self):
        self._cwc(
            where.Where('not f present'),
            [dict(),
             dict(field='f', condition='present',
                  not_phrase=True),
             ])

    def test_not_phrase_not_present(self):
        self._cwc(
            where.Where('not f not present'),
            [dict(),
             dict(field='f', condition='present',
                  not_phrase=True, not_condition=True),
             ])

    def test_not_phrase_eq(self):
        self._cwc(
            where.Where('not f eq b'),
            [dict(),
             dict(field='f', condition='eq', value='b',
                  not_phrase=True),
             ])

    def test_not_phrase_not_eq(self):
        self._cwc(
            where.Where('not f not eq b'),
            [dict(),
             dict(field='f', condition='eq', value='b',
                  not_phrase=True, not_condition=True),
             ])

    def test_not_phrase_num_eq(self):
        self._cwc(
            where.Where('not f num eq b'),
            [dict(),
             dict(field='f', condition='eq', value='b',
                  not_phrase=True, num=True, alpha=False),
             ])

    def test_not_phrase_not_num_eq(self):
        self._cwc(
            where.Where('not f not num eq b'),
            [dict(),
             dict(field='f', condition='eq', value='b',
                  not_phrase=True, not_condition=True, num=True, alpha=False),
             ])

    def test_not_phrase_alpha_eq(self):
        self._cwc(
            where.Where('not f alpha eq b'),
            [dict(),
             dict(field='f', condition='eq', value='b',
                  not_phrase=True, num=False, alpha=True),
             ])

    def test_not_phrase_not_alpha_eq(self):
        self._cwc(
            where.Where('not f not alpha eq b'),
            [dict(),
             dict(field='f', condition='eq', value='b',
                  not_phrase=True, not_condition=True, num=False, alpha=True),
             ])

    def test_not_phrase_ne(self):
        self._cwc(
            where.Where('not f ne b'),
            [dict(),
             dict(field='f', condition='ne', value='b',
                  not_phrase=True),
             ])

    def test_not_phrase_not_ne(self):
        self._cwc(
            where.Where('not f not ne b'),
            [dict(),
             dict(field='f', condition='ne', value='b',
                  not_phrase=True, not_condition=True),
             ])

    def test_not_phrase_num_ne(self):
        self._cwc(
            where.Where('not f num ne b'),
            [dict(),
             dict(field='f', condition='ne', value='b',
                  not_phrase=True, num=True, alpha=False),
             ])

    def test_not_phrase_not_num_ne(self):
        self._cwc(
            where.Where('not f not num ne b'),
            [dict(),
             dict(field='f', condition='ne', value='b',
                  not_phrase=True, not_condition=True, num=True, alpha=False),
             ])

    def test_not_phrase_alpha_ne(self):
        self._cwc(
            where.Where('not f alpha ne b'),
            [dict(),
             dict(field='f', condition='ne', value='b',
                  not_phrase=True, num=False, alpha=True),
             ])

    def test_not_phrase_not_alpha_ne(self):
        self._cwc(
            where.Where('not f not alpha ne b'),
            [dict(),
             dict(field='f', condition='ne', value='b',
                  not_phrase=True, not_condition=True, num=False, alpha=True),
             ])

    def test_not_phrase_gt(self):
        self._cwc(
            where.Where('not f gt b'),
            [dict(),
             dict(field='f', condition='gt', value='b',
                  not_phrase=True),
             ])

    def test_not_phrase_not_gt(self):
        self._cwc(
            where.Where('not f not gt b'),
            [dict(),
             dict(field='f', condition='gt', value='b',
                  not_phrase=True, not_condition=True),
             ])

    def test_not_phrase_num_gt(self):
        self._cwc(
            where.Where('not f num gt b'),
            [dict(),
             dict(field='f', condition='gt', value='b',
                  not_phrase=True, num=True, alpha=False),
             ])

    def test_not_phrase_not_num_gt(self):
        self._cwc(
            where.Where('not f not num gt b'),
            [dict(),
             dict(field='f', condition='gt', value='b',
                  not_phrase=True, not_condition=True, num=True, alpha=False),
             ])

    def test_not_phrase_alpha_gt(self):
        self._cwc(
            where.Where('not f alpha gt b'),
            [dict(),
             dict(field='f', condition='gt', value='b',
                  not_phrase=True, num=False, alpha=True),
             ])

    def test_not_phrase_not_alpha_gt(self):
        self._cwc(
            where.Where('not f not alpha gt b'),
            [dict(),
             dict(field='f', condition='gt', value='b',
                  not_phrase=True, not_condition=True, num=False, alpha=True),
             ])

    def test_not_phrase_lt(self):
        self._cwc(
            where.Where('not f lt b'),
            [dict(),
             dict(field='f', condition='lt', value='b',
                  not_phrase=True),
             ])

    def test_not_phrase_not_lt(self):
        self._cwc(
            where.Where('not f not lt b'),
            [dict(),
             dict(field='f', condition='lt', value='b',
                  not_phrase=True, not_condition=True),
             ])

    def test_not_phrase_num_lt(self):
        self._cwc(
            where.Where('not f num lt b'),
            [dict(),
             dict(field='f', condition='lt', value='b',
                  not_phrase=True, num=True, alpha=False),
             ])

    def test_not_phrase_not_num_lt(self):
        self._cwc(
            where.Where('not f not num lt b'),
            [dict(),
             dict(field='f', condition='lt', value='b',
                  not_phrase=True, not_condition=True, num=True, alpha=False),
             ])

    def test_not_phrase_alpha_lt(self):
        self._cwc(
            where.Where('not f alpha lt b'),
            [dict(),
             dict(field='f', condition='lt', value='b',
                  not_phrase=True, num=False, alpha=True),
             ])

    def test_not_phrase_not_alpha_lt(self):
        self._cwc(
            where.Where('not f not alpha lt b'),
            [dict(),
             dict(field='f', condition='lt', value='b',
                  not_phrase=True, not_condition=True, num=False, alpha=True),
             ])

    def test_not_phrase_le(self):
        self._cwc(
            where.Where('not f le b'),
            [dict(),
             dict(field='f', condition='le', value='b',
                  not_phrase=True),
             ])

    def test_not_phrase_not_le(self):
        self._cwc(
            where.Where('not f not le b'),
            [dict(),
             dict(field='f', condition='le', value='b',
                  not_phrase=True, not_condition=True),
             ])

    def test_not_phrase_num_le(self):
        self._cwc(
            where.Where('not f num le b'),
            [dict(),
             dict(field='f', condition='le', value='b',
                  not_phrase=True, num=True, alpha=False),
             ])

    def test_not_phrase_not_num_le(self):
        self._cwc(
            where.Where('not f not num le b'),
            [dict(),
             dict(field='f', condition='le', value='b',
                  not_phrase=True, not_condition=True, num=True, alpha=False),
             ])

    def test_not_phrase_alpha_le(self):
        self._cwc(
            where.Where('not f alpha le b'),
            [dict(),
             dict(field='f', condition='le', value='b',
                  not_phrase=True, num=False, alpha=True),
             ])

    def test_not_phrase_not_alpha_le(self):
        self._cwc(
            where.Where('not f not alpha le b'),
            [dict(),
             dict(field='f', condition='le', value='b',
                  not_phrase=True, not_condition=True, num=False, alpha=True),
             ])

    def test_not_phrase_ge(self):
        self._cwc(
            where.Where('not f ge b'),
            [dict(),
             dict(field='f', condition='ge', value='b',
                  not_phrase=True),
             ])

    def test_not_phrase_not_ge(self):
        self._cwc(
            where.Where('not f not ge b'),
            [dict(),
             dict(field='f', condition='ge', value='b',
                  not_phrase=True, not_condition=True),
             ])

    def test_not_phrase_num_ge(self):
        self._cwc(
            where.Where('not f num ge b'),
            [dict(),
             dict(field='f', condition='ge', value='b',
                  not_phrase=True, num=True, alpha=False),
             ])

    def test_not_phrase_not_num_ge(self):
        self._cwc(
            where.Where('not f not num ge b'),
            [dict(),
             dict(field='f', condition='ge', value='b',
                  not_phrase=True, not_condition=True, num=True, alpha=False),
             ])

    def test_not_phrase_alpha_ge(self):
        self._cwc(
            where.Where('not f alpha ge b'),
            [dict(),
             dict(field='f', condition='ge', value='b',
                  not_phrase=True, num=False, alpha=True),
             ])

    def test_not_phrase_not_alpha_ge(self):
        self._cwc(
            where.Where('not f not alpha ge b'),
            [dict(),
             dict(field='f', condition='ge', value='b',
                  not_phrase=True, not_condition=True, num=False, alpha=True),
             ])

    def test_not_phrase_before(self):
        self._cwc(
            where.Where('not f before b'),
            [dict(),
             dict(field='f', condition='before', value='b',
                  not_phrase=True),
             ])

    def test_not_phrase_not_before(self):
        self._cwc(
            where.Where('not f not before b'),
            [dict(),
             dict(field='f', condition='before', value='b',
                  not_phrase=True, not_condition=True),
             ])

    def test_not_phrase_num_before(self):
        self._cwc(
            where.Where('not f num before b'),
            [dict(),
             dict(field='f', condition='before', value='b',
                  not_phrase=True, num=True, alpha=False),
             ])

    def test_not_phrase_not_num_before(self):
        self._cwc(
            where.Where('not f not num before b'),
            [dict(),
             dict(field='f', condition='before', value='b',
                  not_phrase=True, not_condition=True, num=True, alpha=False),
             ])

    def test_not_phrase_alpha_before(self):
        self._cwc(
            where.Where('not f alpha before b'),
            [dict(),
             dict(field='f', condition='before', value='b',
                  not_phrase=True, num=False, alpha=True),
             ])

    def test_not_phrase_not_alpha_before(self):
        self._cwc(
            where.Where('not f not alpha before b'),
            [dict(),
             dict(field='f', condition='before', value='b',
                  not_phrase=True, not_condition=True, num=False, alpha=True),
             ])

    def test_not_phrase_after(self):
        self._cwc(
            where.Where('not f after b'),
            [dict(),
             dict(field='f', condition='after', value='b',
                  not_phrase=True),
             ])

    def test_not_phrase_not_after(self):
        self._cwc(
            where.Where('not f not after b'),
            [dict(),
             dict(field='f', condition='after', value='b',
                  not_phrase=True, not_condition=True),
             ])

    def test_not_phrase_num_after(self):
        self._cwc(
            where.Where('not f num after b'),
            [dict(),
             dict(field='f', condition='after', value='b',
                  not_phrase=True, num=True, alpha=False),
             ])

    def test_not_phrase_not_num_after(self):
        self._cwc(
            where.Where('not f not num after b'),
            [dict(),
             dict(field='f', condition='after', value='b',
                  not_phrase=True, not_condition=True, num=True, alpha=False),
             ])

    def test_not_phrase_alpha_after(self):
        self._cwc(
            where.Where('not f alpha after b'),
            [dict(),
             dict(field='f', condition='after', value='b',
                  not_phrase=True, num=False, alpha=True),
             ])

    def test_not_phrase_not_alpha_after(self):
        self._cwc(
            where.Where('not f not alpha after b'),
            [dict(),
             dict(field='f', condition='after', value='b',
                  not_phrase=True, not_condition=True, num=False, alpha=True),
             ])

    def test_not_phrase_above_below(self):
        self._cwc(
            where.Where('not f above c below d'),
            [dict(),
             dict(field='f', condition=('above', 'below'), value=('c', 'd'),
                  not_phrase=True),
             ])

    def test_not_phrase_not_above_below(self):
        self._cwc(
            where.Where('not f not above c below d'),
            [dict(),
             dict(field='f', condition=('above', 'below'), value=('c', 'd'),
                  not_phrase=True, not_condition=True),
             ])

    def test_not_phrase_num_above_below(self):
        self._cwc(
            where.Where('not f num above c below d'),
            [dict(),
             dict(field='f', condition=('above', 'below'), value=('c', 'd'),
                  not_phrase=True, num=True, alpha=False),
             ])

    def test_not_phrase_not_num_above_below(self):
        self._cwc(
            where.Where('not f not num above c below d'),
            [dict(),
             dict(field='f', condition=('above', 'below'), value=('c', 'd'),
                  not_phrase=True, not_condition=True, num=True, alpha=False),
             ])

    def test_not_phrase_alpha_above_below(self):
        self._cwc(
            where.Where('not f alpha above c below d'),
            [dict(),
             dict(field='f', condition=('above', 'below'), value=('c', 'd'),
                  not_phrase=True, num=False, alpha=True),
             ])

    def test_not_phrase_not_alpha_above_below(self):
        self._cwc(
            where.Where('not f not alpha above c below d'),
            [dict(),
             dict(field='f', condition=('above', 'below'), value=('c', 'd'),
                  not_phrase=True, not_condition=True, num=False, alpha=True),
             ])

    def test_not_phrase_above_to(self):
        self._cwc(
            where.Where('not f above c to d'),
            [dict(),
             dict(field='f', condition=('above', 'to'), value=('c', 'd'),
                  not_phrase=True),
             ])

    def test_not_phrase_not_above_to(self):
        self._cwc(
            where.Where('not f not above c to d'),
            [dict(),
             dict(field='f', condition=('above', 'to'), value=('c', 'd'),
                  not_phrase=True, not_condition=True),
             ])

    def test_not_phrase_num_above_to(self):
        self._cwc(
            where.Where('not f num above c to d'),
            [dict(),
             dict(field='f', condition=('above', 'to'), value=('c', 'd'),
                  not_phrase=True, num=True, alpha=False),
             ])

    def test_not_phrase_not_num_above_to(self):
        self._cwc(
            where.Where('not f not num above c to d'),
            [dict(),
             dict(field='f', condition=('above', 'to'), value=('c', 'd'),
                  not_phrase=True, not_condition=True, num=True, alpha=False),
             ])

    def test_not_phrase_alpha_above_to(self):
        self._cwc(
            where.Where('not f alpha above c to d'),
            [dict(),
             dict(field='f', condition=('above', 'to'), value=('c', 'd'),
                  not_phrase=True, num=False, alpha=True),
             ])

    def test_not_phrase_not_alpha_above_to(self):
        self._cwc(
            where.Where('not f not alpha above c to d'),
            [dict(),
             dict(field='f', condition=('above', 'to'), value=('c', 'd'),
                  not_phrase=True, not_condition=True, num=False, alpha=True),
             ])

    def test_not_phrase_from_below(self):
        self._cwc(
            where.Where('not f from c below d'),
            [dict(),
             dict(field='f', condition=('from', 'below'), value=('c', 'd'),
                  not_phrase=True),
             ])

    def test_not_phrase_not_from_below(self):
        self._cwc(
            where.Where('not f not from c below d'),
            [dict(),
             dict(field='f', condition=('from', 'below'), value=('c', 'd'),
                  not_phrase=True, not_condition=True),
             ])

    def test_not_phrase_num_from_below(self):
        self._cwc(
            where.Where('not f num from c below d'),
            [dict(),
             dict(field='f', condition=('from', 'below'), value=('c', 'd'),
                  not_phrase=True, num=True, alpha=False),
             ])

    def test_not_phrase_not_num_from_below(self):
        self._cwc(
            where.Where('not f not num from c below d'),
            [dict(),
             dict(field='f', condition=('from', 'below'), value=('c', 'd'),
                  not_phrase=True, not_condition=True, num=True, alpha=False),
             ])

    def test_not_phrase_alpha_from_below(self):
        self._cwc(
            where.Where('not f alpha from c below d'),
            [dict(),
             dict(field='f', condition=('from', 'below'), value=('c', 'd'),
                  not_phrase=True, num=False, alpha=True),
             ])

    def test_not_phrase_not_alpha_from_below(self):
        self._cwc(
            where.Where('not f not alpha from c below d'),
            [dict(),
             dict(field='f', condition=('from', 'below'), value=('c', 'd'),
                  not_phrase=True, not_condition=True, num=False, alpha=True),
             ])

    def test_not_phrase_from_to(self):
        self._cwc(
            where.Where('not f from c to d'),
            [dict(),
             dict(field='f', condition=('from', 'to'), value=('c', 'd'),
                  not_phrase=True),
             ])

    def test_not_phrase_not_from_to(self):
        self._cwc(
            where.Where('not f not from c to d'),
            [dict(),
             dict(field='f', condition=('from', 'to'), value=('c', 'd'),
                  not_phrase=True, not_condition=True),
             ])

    def test_not_phrase_num_from_to(self):
        self._cwc(
            where.Where('not f num from c to d'),
            [dict(),
             dict(field='f', condition=('from', 'to'), value=('c', 'd'),
                  not_phrase=True, num=True, alpha=False),
             ])

    def test_not_phrase_not_num_from_to(self):
        self._cwc(
            where.Where('not f not num from c to d'),
            [dict(),
             dict(field='f', condition=('from', 'to'), value=('c', 'd'),
                  not_phrase=True, not_condition=True, num=True, alpha=False),
             ])

    def test_not_phrase_alpha_from_to(self):
        self._cwc(
            where.Where('not f alpha from c to d'),
            [dict(),
             dict(field='f', condition=('from', 'to'), value=('c', 'd'),
                  not_phrase=True, num=False, alpha=True),
             ])

    def test_not_phrase_not_alpha_from_to(self):
        self._cwc(
            where.Where('not f not alpha from c to d'),
            [dict(),
             dict(field='f', condition=('from', 'to'), value=('c', 'd'),
                  not_phrase=True, not_condition=True, num=False, alpha=True),
             ])

    def _cwc(
        self,
        where_clause,
        expected_clauses):
        where_clause.lex()
        where_clause.parse()
        ae = self.assertEqual
        for e, c in enumerate(
            where_clause.node.get_clauses_from_root_in_walk_order()):
            wc = where.WhereClause()
            wc.__dict__.update(expected_clauses[e])
            ae(wc.operator, c.operator)
            ae(wc.field, c.field)
            ae(wc.condition, c.condition)
            ae(wc.value, c.value)
            ae(wc.not_phrase, c.not_phrase)
            ae(wc.not_condition, c.not_condition)
            ae(wc.not_value, c.not_value)
            ae(wc.num, c.num)
            ae(wc.alpha, c.alpha)


class Where_parse_multi_not_phrasesTC(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_not_first_of_two_phrases(self):
        self._cwc(
            where.Where('not f is a and f is b'),
            [dict(),
             dict(field='f', condition='is', value='a', not_phrase=True),
             dict(field='f', condition='is', value='b', operator='and'),
             ])

    def test_multiuse_not_first_of_two_phrases(self):
        self._cwc(
            where.Where('not f is not a and f is b'),
            [dict(),
             dict(field='f', condition='is', value='a',
                  not_phrase=True, not_value=True),
             dict(field='f', condition='is', value='b', operator='and'),
             ])
        self._cwc(
            where.Where('not f is not a and f is not b'),
            [dict(),
             dict(field='f', condition='is', value='a',
                  not_phrase=True, not_value=True),
             dict(field='f', condition='is', value='b', operator='and',
                  not_value=True),
             ])
        self._cwc(
            where.Where('not f is a and f is not b'),
            [dict(),
             dict(field='f', condition='is', value='a',
                  not_phrase=True),
             dict(field='f', condition='is', value='b', operator='and',
                  not_value=True),
             ])

    def test_not_both_of_two_phrases(self):
        self._cwc(
            where.Where('not f is a and not f is b'),
            [dict(),
             dict(field='f', condition='is', value='a', not_phrase=True),
             dict(field='f', condition='is', value='b', operator='and',
                  not_phrase=True),
             ])

    def test_multiuse_not_both_of_two_phrases(self):
        self._cwc(
            where.Where('not f is not a and not f is b'),
            [dict(),
             dict(field='f', condition='is', value='a',
                  not_phrase=True, not_value=True),
             dict(field='f', condition='is', value='b', operator='and',
                  not_phrase=True),
             ])
        self._cwc(
            where.Where('not f is not a and not f is not b'),
            [dict(),
             dict(field='f', condition='is', value='a',
                  not_phrase=True, not_value=True),
             dict(field='f', condition='is', value='b', operator='and',
                  not_phrase=True, not_value=True),
             ])
        self._cwc(
            where.Where('not f is a and not f is not b'),
            [dict(),
             dict(field='f', condition='is', value='a',
                  not_phrase=True),
             dict(field='f', condition='is', value='b', operator='and',
                  not_phrase=True, not_value=True),
             ])

    def test_all_five_simple_positive_not_phrases(self):
        self._cwc(
            where.Where(
                ' '.join(('not f is a or not f like b or not f present or',
                          'not f eq c or not f from d to e or not f is f',
                          ))),
            [dict(),
             dict(field='f', condition='is', value='a',
                  not_phrase=True),
             dict(field='f', condition='like', value='b',
                  operator='or', not_phrase=True),
             dict(field='f', condition='present',
                  operator='or', not_phrase=True),
             dict(field='f', condition='eq', value='c',
                  operator='or', not_phrase=True),
             dict(field='f',
                  condition=('from', 'to'), value=('d', 'e'),
                  operator='or', not_phrase=True),
             dict(field='f', condition='is', value='f',
                  operator='or', not_phrase=True),
             ])

    def test_all_five_simple_negative_not_phrases(self):
        self._cwc(
            where.Where(
                ' '.join(('not f is not a or not f not like b or',
                          'not f not present or not f not eq c or',
                          'not f not from d to e or not f is not f',
                          ))),
            [dict(),
             dict(field='f', condition='is', value='a',
                  not_value=True, not_phrase=True),
             dict(field='f', condition='like', value='b',
                  operator='or', not_condition=True, not_phrase=True),
             dict(field='f', condition='present',
                  operator='or', not_condition=True, not_phrase=True),
             dict(field='f', condition='eq', value='c',
                  operator='or', not_condition=True, not_phrase=True),
             dict(field='f',
                  condition=('from', 'to'), value=('d', 'e'),
                  operator='or', not_condition=True, not_phrase=True),
             dict(field='f', condition='is', value='f',
                  operator='or', not_value=True, not_phrase=True),
             ])

    def test_all_five_simple_positive_num_not_phrases(self):
        self._cwc(
            where.Where(
                ' '.join(('not f is a or not f like b or',
                          'not f present or not f num eq c or',
                          'not f num from d to e or not f is f',
                          ))),
            [dict(),
             dict(field='f', condition='is', value='a',
                  not_phrase=True),
             dict(field='f', condition='like', value='b',
                  operator='or', not_phrase=True),
             dict(field='f', condition='present',
                  operator='or', not_phrase=True),
             dict(field='f', condition='eq', value='c', operator='or',
                  num=True, alpha=False, not_phrase=True),
             dict(field='f',
                  condition=('from', 'to'), value=('d', 'e'),
                  operator='or', num=True, alpha=False, not_phrase=True),
             dict(field='f', condition='is', value='f',
                  operator='or', not_phrase=True),
             ])

    def test_all_five_simple_negative_num_not_phrases(self):
        self._cwc(
            where.Where(
                ' '.join(('not f is not a or not f not like b or',
                          'not f not present or not f not num eq c or',
                          'not f not num from d to e or not f is not f',
                          ))),
            [dict(),
             dict(field='f', condition='is', value='a',
                  not_value=True, not_phrase=True),
             dict(field='f', condition='like', value='b',
                  operator='or', not_condition=True, not_phrase=True),
             dict(field='f', condition='present',
                  operator='or', not_condition=True, not_phrase=True),
             dict(field='f', condition='eq', value='c',
                  operator='or', not_condition=True, not_phrase=True,
                  num=True, alpha=False),
             dict(field='f',
                  condition=('from', 'to'), value=('d', 'e'),
                  operator='or', not_condition=True, not_phrase=True,
                  num=True, alpha=False),
             dict(field='f', condition='is', value='f',
                  operator='or', not_value=True, not_phrase=True),
             ])

    def test_all_five_simple_positive_alpha_not_phrases(self):
        self._cwc(
            where.Where(
                ' '.join(('not f is a or not f like b or not f present or',
                          'not f alpha eq c or not f alpha from d to e or',
                          'not f is f',
                          ))),
            [dict(),
             dict(field='f', condition='is', value='a',
                  not_phrase=True),
             dict(field='f', condition='like', value='b',
                  operator='or', not_phrase=True),
             dict(field='f', condition='present',
                  operator='or', not_phrase=True),
             dict(field='f', condition='eq', value='c', operator='or',
                  num=False, alpha=True, not_phrase=True),
             dict(field='f',
                  condition=('from', 'to'), value=('d', 'e'),
                  operator='or', num=False, alpha=True, not_phrase=True),
             dict(field='f', condition='is', value='f',
                  operator='or', not_phrase=True),
             ])

    def test_all_five_simple_negative_alpha_not_phrases(self):
        self._cwc(
            where.Where(
                ' '.join(('not f is not a or not f not like b or',
                          'not f not present or not f not alpha eq c or',
                          'not f not alpha from d to e or not f is not f',
                          ))),
            [dict(),
             dict(field='f', condition='is', value='a',
                  not_value=True, not_phrase=True),
             dict(field='f', condition='like', value='b',
                  operator='or', not_condition=True, not_phrase=True),
             dict(field='f', condition='present',
                  operator='or', not_condition=True, not_phrase=True),
             dict(field='f', condition='eq', value='c',
                  operator='or', not_condition=True, not_phrase=True,
                  num=False, alpha=True),
             dict(field='f',
                  condition=('from', 'to'), value=('d', 'e'),
                  operator='or', not_condition=True, not_phrase=True,
                  num=False, alpha=True),
             dict(field='f', condition='is', value='f',
                  operator='or', not_value=True, not_phrase=True),
             ])

    def test_all_three_simple_boolean_not_phrases(self):
        self._cwc(
            where.Where(
                ' '.join(('not f is a or not f like b or not f present and',
                          'not f eq c and not f from d to e nor',
                          'not f is f nor not f le g',
                          ))),
            [dict(),
             dict(field='f', condition='is', value='a',
                  not_phrase=True),
             dict(field='f', condition='like', value='b',
                  operator='or', not_phrase=True),
             dict(field='f', condition='present',
                  operator='or', not_phrase=True),
             dict(field='f', condition='eq', value='c',
                  operator='and', not_phrase=True),
             dict(field='f',
                  condition=('from', 'to'), value=('d', 'e'),
                  operator='and', not_phrase=True),
             dict(field='f', condition='is', value='f',
                  operator='nor', not_phrase=True),
             dict(field='f', condition='le', value='g',
                  operator='nor', not_phrase=True),
             ])

    def _cwc(
        self,
        where_clause,
        expected_clauses):
        where_clause.lex()
        where_clause.parse()
        ae = self.assertEqual
        for e, c in enumerate(
            where_clause.node.get_clauses_from_root_in_walk_order()):
            wc = where.WhereClause()
            wc.__dict__.update(expected_clauses[e])
            ae(wc.operator, c.operator)
            ae(wc.field, c.field)
            ae(wc.condition, c.condition)
            ae(wc.value, c.value)
            ae(wc.not_phrase, c.not_phrase)
            ae(wc.not_condition, c.not_condition)
            ae(wc.not_value, c.not_value)
            ae(wc.num, c.num)
            ae(wc.alpha, c.alpha)


class Where_parse_parenthesis_phrasesTC(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_parenthesis_is(self):
        self._cwc(
            where.Where('(f is a)'),
            [dict(),
             dict(),
             dict(field='f', condition='is', value='a'),
             ])
        self._cwc(
            where.Where('( f is a)'),
            [dict(),
             dict(),
             dict(field='f', condition='is', value='a'),
             ])
        self._cwc(
            where.Where('( f is a )'),
            [dict(),
             dict(),
             dict(field='f', condition='is', value='a'),
             ])
        self._cwc(
            where.Where('(f is a )'),
            [dict(),
             dict(),
             dict(field='f', condition='is', value='a'),
             ])
        self._cwc(
            where.Where(' ( f is a ) '),
            [dict(),
             dict(),
             dict(field='f', condition='is', value='a'),
             ])

    def test_parenthesis_is_two(self):
        node = self._cwc(
            where.Where('(f is a or f is b) and (g is c or g is d)'),
            [dict(),
             dict(),
             dict(field='f', condition='is', value='a'),
             dict(field='f', condition='is', value='b', operator='or'),
             dict(operator='and'),
             dict(field='g', condition='is', value='c'),
             dict(field='g', condition='is', value='d', operator='or'),
             ])
        for e, n in enumerate(
            [dict(down=node[1]),
             dict(right=node[4], down=node[2], up=node[0]),
             dict(right=node[3], up=node[1]),
             dict(left=node[2], up=node[1]),
             dict(left=node[1], down=node[5], up=node[0]),
             dict(right=node[6], up=node[4]),
             dict(left=node[5], up=node[4]),
             ]):
            self._cws(node[e], n)

    def test_parenthesis_is_three(self):
        node = self._cwc(
            where.Where(' '.join(
                ('(f is a or f is b) and',
                 '(g is c or g is d) and',
                 '(h is e or h is i)'))),
            [dict(),
             dict(),
             dict(field='f', condition='is', value='a'),
             dict(field='f', condition='is', value='b', operator='or'),
             dict(operator='and'),
             dict(field='g', condition='is', value='c'),
             dict(field='g', condition='is', value='d', operator='or'),
             dict(operator='and'),
             dict(field='h', condition='is', value='e'),
             dict(field='h', condition='is', value='i', operator='or'),
             ])
        for e, n in enumerate(
            [dict(down=node[1]),
             dict(right=node[4], down=node[2], up=node[0]),
             dict(right=node[3], up=node[1]),
             dict(left=node[2], up=node[1]),
             dict(left=node[1], right=node[7], down=node[5], up=node[0]),
             dict(right=node[6], up=node[4]),
             dict(left=node[5], up=node[4]),
             dict(left=node[4], down=node[8], up=node[0]),
             dict(right=node[9], up=node[7]),
             dict(left=node[8], up=node[7]),
             ]):
            self._cws(node[e], n)

    def test_parenthesis_is_three_nested(self):
        node = self._cwc(
            where.Where(' '.join(
                ('(f is a nor (w gt y or x lt z) or f is b) and',
                 '(g is c nor (not w gt y nor x lt z) or g is d) and',
                 '(h is e nor (w not gt y or not x lt z) or p is q)'))),
            [dict(),
             dict(),
             dict(field='f', condition='is', value='a'),
             dict(operator='nor'),
             dict(field='w', condition='gt', value='y'),
             dict(field='x', condition='lt', value='z', operator='or'),
             dict(field='f', condition='is', value='b', operator='or'),
             dict(operator='and'),
             dict(field='g', condition='is', value='c'),
             dict(operator='nor'),
             dict(field='w', condition='gt', value='y', not_phrase=True),
             dict(field='x', condition='lt', value='z', operator='nor'),
             dict(field='g', condition='is', value='d', operator='or'),
             dict(operator='and'),
             dict(field='h', condition='is', value='e'),
             dict(operator='nor'),
             dict(field='w', condition='gt', value='y', not_condition=True),
             dict(field='x', condition='lt', value='z', operator='or',
                  not_phrase=True),
             dict(field='p', condition='is', value='q', operator='or'),
             ])
        for e, n in enumerate(
            [dict(down=node[1]),
             dict(right=node[7], down=node[2], up=node[0]),
             dict(right=node[3], up=node[1]),
             dict(left=node[2], right=node[6], up=node[1], down=node[4]),
             dict(right=node[5], up=node[3]),
             dict(left=node[4], up=node[3]),
             dict(left=node[3], up=node[1]),
             dict(left=node[1], right=node[13], down=node[8], up=node[0]),
             dict(right=node[9], up=node[7]),
             dict(left=node[8], right=node[12], up=node[7], down=node[10]),
             dict(right=node[11], up=node[9]),
             dict(left=node[10], up=node[9]),
             dict(left=node[9], up=node[7]),
             dict(left=node[7], down=node[14], up=node[0]),
             dict(right=node[15], up=node[13]),
             dict(left=node[14], right=node[18], up=node[13],
                  down=node[16]),
             dict(right=node[17], up=node[15]),
             dict(left=node[16], up=node[15]),
             dict(left=node[15], up=node[13]),
             ]):
            self._cws(node[e], n)

    def test_parenthesis_like(self):
        self._cwc(
            where.Where('(f like a)'),
            [dict(),
             dict(),
             dict(field='f', condition='like', value='a'),
             ])

    def test_parenthesis_present(self):
        self._cwc(
            where.Where('(f present)'),
            [dict(),
             dict(),
             dict(field='f', condition='present'),
             ])

    def test_parenthesis_after(self):
        self._cwc(
            where.Where('(f after a)'),
            [dict(),
             dict(),
             dict(field='f', condition='after', value='a'),
             ])

    def test_parenthesis_below(self):
        self._cwc(
            where.Where('(f from a below b)'),
            [dict(),
             dict(),
             dict(field='f', condition=('from', 'below'), value=('a', 'b')),
             ])

    def test_nested_parenthesis(self):
        self._cwc(
            where.Where('((f is a))'),
            [dict(),
             dict(),
             dict(),
             dict(field='f', condition='is', value='a'),
             ])
        self._cwc(
            where.Where('(((f is a)))'),
            [dict(),
             dict(),
             dict(),
             dict(),
             dict(field='f', condition='is', value='a'),
             ])

    def _cwc(
        self,
        where_clause,
        expected_clauses):
        where_clause.lex()
        where_clause.parse()
        ae = self.assertEqual
        node = where_clause.node.get_clauses_from_root_in_walk_order()
        for e, c in enumerate(node):
            wc = where.WhereClause()
            wc.__dict__.update(expected_clauses[e])
            ae(wc.operator, c.operator)
            ae(wc.field, c.field)
            ae(wc.condition, c.condition)
            ae(wc.value, c.value)
            ae(wc.not_phrase, c.not_phrase)
            ae(wc.not_condition, c.not_condition)
            ae(wc.not_value, c.not_value)
            ae(wc.num, c.num)
            ae(wc.alpha, c.alpha)
        return node

    def _cws(self, where_phrase, expected_navigation):
        wc = where.WhereClause()
        wc.__dict__.update(expected_navigation)
        ae = self.assertEqual
        ae(where_phrase.left, wc.left)
        ae(where_phrase.right, wc.right)
        ae(where_phrase.up, wc.up)
        ae(where_phrase.down, wc.down)


class Where_errorTC(unittest.TestCase):

    def setUp(self):
        self.w = where.Where('')

    def tearDown(self):
        pass

    def test____assumptions(self):
        w = self.w
        self.assertEqual(set((w.__dict__)),
                         set(('present_dispatch',
                              'range_start_dispatch',
                              'range_end_dispatch',
                              'boolean_dispatch',
                              'invert_dispatch',
                              'compare_dispatch',
                              'is_dispatch',
                              'like_dispatch',
                              'left_parenthesis_dispatch',
                              'right_parenthesis_dispatch',
                              'value_dispatch',
                              'alphanumeric_dispatch',
                              'statement',
                              'node',
                              'token_dispatch',
                              '_processors',
                              'tokens',
                              '_f_or_v',
                              'none_dispatch',
                              '_not',
                              )))

    def test____unknown_state(self):
        w = self.w
        lookups = (w.present,
                   w.range_start,
                   w.range_end,
                   w.boolean,
                   w.invert,
                   w.compare,
                   w.left_parenthesis,
                   w.right_parenthesis,
                   w.value,
                   w.alphanumeric,
                   )
        for d in lookups:
            self.assertRaisesRegex(
                where.WhereError,
                'Unknown state',
                d,
                *(lookups, ''))

    def test_report(self):
        self.assertRaisesRegex(
            where.WhereError,
            'Token not allowed in state',
            self.w.report,
            *('',))

    def test_error(self):
        self.assertRaisesRegex(
            where.WhereError,
            'Unknown state',
            self.w.error,
            *('',))


class Where_evaluate_traceTC(unittest.TestCase):

    # Some of this looks too complex to be a test, or the test needs testing!

    # Many arbitrary 'c0'-like values in self._tec() arguments changed to
    # set('c0')-like values to keep evaluate() calls happy after addition of
    # proper constraints code.  This test does not get as far as expecting
    # sensible answers in this respect.

    def setUp(self):
        self.clauses = {}
        self.trace = []

    def tearDown(self):
        pass

    def test_evaluate_is(self):
        query = 'f is a'
        a = [dict(),
             dict(field='f', condition='is', value='a'),
             ]
        for e, c in enumerate(self._evaluate(query)):
            self._cwc(c, a[e])
        self._tec((('is', 1, 'c0'),
                   ('answer', 1, 0, set('c0')),
                   ))

    def test_evaluate_parenthesis_is(self):
        query = '(f is a)'
        a = [dict(),
             dict(),
             dict(field='f', condition='is', value='a'),
             ]
        for e, c in enumerate(self._evaluate(query)):
            self._cwc(c, a[e])
        self._tec((('is', 2, 'c0'),
                   ('answer', 2, 1, set('c0')),
                   ('answer', 1, 0, set('c0')),
                   ))

    def test_evaluate_sample_multiple(self):
        query = ' '.join(
            ('(f is a nor (w gt y or x lt z) or f is b) and',
             '(g is c nor (not w gt y nor x lt z) or g is d) and',
             '(h is e nor (w not gt y or not x lt z) or h is i)'))
        a = [dict(),
             dict(),
             dict(field='f', condition='is', value='a'),
             dict(operator='nor'),
             dict(field='w', condition='gt', value='y'),
             dict(field='x', condition='lt', value='z', operator='or'),
             dict(field='f', condition='is', value='b', operator='or'),
             dict(operator='and'),
             dict(field='g', condition='is', value='c'),
             dict(operator='nor'),
             dict(field='w', condition='gt', value='y', not_phrase=True),
             dict(field='x', condition='lt', value='z', operator='nor'),
             dict(field='g', condition='is', value='d', operator='or'),
             dict(operator='and'),
             dict(field='h', condition='is', value='e'),
             dict(operator='nor'),
             dict(field='w', condition='gt', value='y', not_condition=True),
             dict(field='x', condition='lt', value='z', operator='or',
                  not_phrase=True),
             dict(field='h', condition='is', value='i', operator='or'),
             ]
        for e, c in enumerate(self._evaluate(query)):
            self._cwc(c, a[e])
        self._tec((('is', 2, 'c0'),
                   ('gt', 4, 'c1'),
                   ('lt', 5, 'c2'),
                   ('or', 4, 5, set('c1'), set('c2')),
                   ('answer', 5, 3, set('c1')),
                   ('is', 6, 'c3'),
                   ('nor', 2, 3, set('c0'), set('c1')),
                   ('or', 3, 6, set('c0'), set('c3')),
                   ('answer', 6, 1, set('c0')),
                   ('is', 8, 'c4'),
                   ('gt', 10, 'c5'),
                   ('lt', 11, 'c6'),
                   ('nor', 10, 11, set('c5'), set('c6')),
                   ('answer', 11, 9, set('c5')),
                   ('is', 12, 'c7'),
                   ('nor', 8, 9, set('c4'), set('c5')),
                   ('or', 9, 12, set('c4'), set('c7')),
                   ('answer', 12, 7, set('c4')),
                   ('is', 14, 'c8'),
                   ('gt', 16, 'c9'),
                   ('lt', 17, 'c10'),
                   ('or', 16, 17, set('c9'), set('c10')),
                   ('answer', 17, 15, set('c9')),
                   ('is', 18, 'c11'),
                   ('nor', 14, 15, set('c8'), set('c9')),
                   ('or', 15, 18, set('c8'), set('c11')),
                   ('answer', 18, 13, set('c8')),
                   ('and', 1, 7, set('c0'), set('c4')),
                   ('and', 7, 13, set('c0'), set('c8')),
                   ('answer', 13, 0, set('c0')),
                   ))

    def test_evaluate_quotes_1(self):
        query = ' '.join(
            ('''"fr'ed or "  sand'  "  'mary  above "5" below "10"''',
             '''and("x" eq "y")''',
             ))
        a = [dict(),
             dict(field='''fr'ed or sand  "  mary''',
                  condition=('above', 'below'),
                  value=('5', '10')),
             dict(operator='and'),
             dict(field='x', condition='eq', value='y'),
             ]
        for e, c in enumerate(self._evaluate(query)):
            self._cwc(c, a[e])
        self._tec(((('above', 'below'), 1, 'c0'),
                   ('eq', 3, 'c1'),
                   ('answer', 3, 2, set('c1')),
                   ('and', 1, 2, set('c0'), set('c1')),
                   ('answer', 2, 0, set('c0')),
                   ))

    def test_evaluate_quotes_2(self):
        query = ' '.join(
            ('''"fr'ed or "  sand'  "  'mary  above "5" below "10"''',
             '''and "x" eq "y"''',
             ))
        a = [dict(),
             dict(field='''fr'ed or sand  "  mary''',
                  condition=('above', 'below'),
                  value=('5', '10')),
             dict(field='x', condition='eq', value='y', operator='and'),
             ]
        for e, c in enumerate(self._evaluate(query)):
            self._cwc(c, a[e])
        self._tec(((('above', 'below'), 1, 'c0'),
                   ('eq', 2, 'c1'),
                   ('and', 1, 2, set('c0'), set('c1')),
                   ('answer', 2, 0, set('c0')),
                   ))

    def test_evaluate_quotes_3(self):
        # See _evaluate() and Processors.not_condition(): this test did not
        # change when the delayed processing of non-index conditions was
        # introduced to get a single pass through the data records.
        query = ' '.join(
            ('''"fr'ed or "  sand'  "  'mary  above "5" below "10"''',
             '''and(x present)''',
             ))
        a = [dict(),
             dict(field='''fr'ed or sand  "  mary''',
                  condition=('above', 'below'),
                  value=('5', '10')),
             dict(operator='and'),
             dict(field='x', condition='present'),
             ]
        for e, c in enumerate(self._evaluate(query)):
            self._cwc(c, a[e])
        self._tec(((('above', 'below'), 1, 'c0'),
                   ('present', 3, 'c1'),
                   ('answer', 3, 2, set('c1')),
                   ('and', 1, 2, set('c0'), set('c1')),
                   ('answer', 2, 0, set('c0')),
                   ))

    def test_evaluate_quotes_4(self):
        query = ' '.join(
            ("meal eq fish 'and' chips",
             ))
        a = [dict(),
             dict(field='meal',
                  condition=('eq'),
                  value=('fishandchips')),
             ]
        for e, c in enumerate(self._evaluate(query)):
            self._cwc(c, a[e])
        self._tec(((('eq'), 1, 'c0'),
                   ('answer', 1, 0, set('c0')),
                   ))

    def test_evaluate_quotes_5(self):
        query = ' '.join(
            ("meal eq fish' and 'chips",
             ))
        a = [dict(),
             dict(field='meal',
                  condition=('eq'),
                  value=('fish and chips')),
             ]
        for e, c in enumerate(self._evaluate(query)):
            self._cwc(c, a[e])
        self._tec(((('eq'), 1, 'c0'),
                   ('answer', 1, 0, set('c0')),
                   ))

    def test_evaluate_quotes_6(self):
        query = ' '.join(
            ("meal eq 'fish and chips'",
             ))
        a = [dict(),
             dict(field='meal',
                  condition=('eq'),
                  value=('fish and chips')),
             ]
        for e, c in enumerate(self._evaluate(query)):
            self._cwc(c, a[e])
        self._tec(((('eq'), 1, 'c0'),
                   ('answer', 1, 0, set('c0')),
                   ))

    def test_evaluate_quotes_7(self):
        query = ' '.join(
            ("name eq O'Neill",
             ))
        a = [dict(),
             dict(field='name',
                  condition=('eq'),
                  value=("O'Neill")),
             ]
        for e, c in enumerate(self._evaluate(query)):
            self._cwc(c, a[e])
        self._tec(((('eq'), 1, 'c0'),
                   ('answer', 1, 0, set('c0')),
                   ))

    def test_evaluate_multiple_boolean_1(self):
        query = 'f1 eq v1 or f2 eq v2 or f3 eq v3 or f4 eq v4 or f5 after v5'
        a = [dict(),
             dict(field='f1', condition='eq', value='v1'),
             dict(field='f2', condition='eq', value='v2', operator='or'),
             dict(field='f3', condition='eq', value='v3', operator='or'),
             dict(field='f4', condition='eq', value='v4', operator='or'),
             dict(field='f5', condition='after', value='v5', operator='or'),
             ]
        for e, c in enumerate(self._evaluate(query)):
            self._cwc(c, a[e])
        self._tec((('eq', 1, 'c0'),
                   ('eq', 2, 'c1'),
                   ('eq', 3, 'c2'),
                   ('eq', 4, 'c3'),
                   ('after', 5, 'c4'),
                   ('or', 1, 2, set('c0'), set('c1')),
                   ('or', 2, 3, set('c0'), set('c2')),
                   ('or', 3, 4, set('c0'), set('c3')),
                   ('or', 4, 5, set('c0'), set('c4')),
                   ('answer', 5, 0, set('c0')),
                   ))

    def test_evaluate_multiple_boolean_2(self):
        query = 'f1 eq v1 or f2 eq v2 or f3 eq v3 and f4 eq v4 or f5 after v5'
        a = [dict(),
             dict(field='f1', condition='eq', value='v1'),
             dict(field='f2', condition='eq', value='v2', operator='or'),
             dict(field='f3', condition='eq', value='v3', operator='or'),
             dict(field='f4', condition='eq', value='v4', operator='and'),
             dict(field='f5', condition='after', value='v5', operator='or'),
             ]
        for e, c in enumerate(self._evaluate(query)):
            self._cwc(c, a[e])
        self._tec((('eq', 1, 'c0'),
                   ('eq', 2, 'c1'),
                   ('eq', 3, 'c2'),
                   ('eq', 4, 'c3'),
                   ('after', 5, 'c4'),
                   ('and', 3, 4, set('c2'), set('c3')),
                   ('or', 1, 2, set('c0'), set('c1')),
                   ('or', 2, 3, set('c0'), set('c2')),
                   ('or', 4, 5, set('c2'), set('c4')),
                   ('answer', 5, 0, set('c2')),
                   ))

    def test_evaluate_multiple_boolean_3(self):
        query = 'f1 eq v1 or f2 eq v2 or (f3 eq v3 and f4 eq v4) or f5 after v5'
        a = [dict(),
             dict(field='f1', condition='eq', value='v1'),
             dict(field='f2', condition='eq', value='v2', operator='or'),
             dict(operator='or'),
             dict(field='f3', condition='eq', value='v3'),
             dict(field='f4', condition='eq', value='v4', operator='and'),
             dict(field='f5', condition='after', value='v5', operator='or'),
             ]
        for e, c in enumerate(self._evaluate(query)):
            self._cwc(c, a[e])
        self._tec((('eq', 1, 'c0'),
                   ('eq', 2, 'c1'),
                   ('eq', 4, 'c2'),
                   ('eq', 5, 'c3'),
                   ('and', 4, 5, set('c2'), set('c3')),
                   ('answer', 5, 3, set('c2')),
                   ('after', 6, 'c4'),
                   ('or', 1, 2, set('c0'), set('c1')),
                   ('or', 2, 3, set('c0'), set('c2')),
                   ('or', 3, 6, set('c0'), set('c4')),
                   ('answer', 6, 0, set('c0')),
                   ))

    def test_evaluate_nested_parentheses_01(self):
        query = '(f is a)'
        a = [dict(),
             dict(),
             dict(field='f', condition='is', value='a'),
             ]
        for e, c in enumerate(self._evaluate(query)):
            self._cwc(c, a[e])
        self._tec((('is', 2, 'c0'),
                   ('answer', 2, 1, set('c0')),
                   ('answer', 1, 0, set('c0')),
                   ))

    def test_evaluate_nested_parentheses_02(self):
        query = '((f is a))'
        a = [dict(),
             dict(),
             dict(),
             dict(field='f', condition='is', value='a'),
             ]
        for e, c in enumerate(self._evaluate(query)):
            self._cwc(c, a[e])
        self._tec((('is', 3, 'c0'),
                   ('answer', 3, 2, set('c0')),
                   ('answer', 2, 1, set('c0')),
                   ('answer', 1, 0, set('c0')),
                   ))

    def _evaluate(self, query):

        class Processors():

            def __init__(self, trace, clauses):
                self.trace = trace
                self.clauses = clauses
                self.evaluations = {}
                self.existence = {}
                
            def condition(self, obj):
                c = ''.join(('c', str(len(self.evaluations))))
                self.trace.append(
                    (obj.condition,
                     self.clauses[hex(id(obj))],
                     c))
                self.evaluations[obj.get_condition()] = c
                obj.result.answer = set(c)
            
            def operator(self, obj):
                ce = self.clauses
                self.trace.append(
                    (obj.operator,
                     ce[hex(id(obj.left))],
                     ce[hex(id(obj))],
                     obj.left.result.answer,
                     obj.result.answer))
                obj.result = obj.left.result
            
            def answer(self, obj):
                ce = self.clauses
                self.trace.append(
                    ('answer',
                     ce[hex(id(obj))],
                     ce[hex(id(obj.up))],
                     obj.result.answer))
                obj.up.result = obj.result
            
            def initialize_answer(self, obj):
                obj.result.answer = set()
            
            def not_condition(self, obj):
                # Do the trace which was not done in the index processing.
                self.condition(obj)
            
            def get_existence(self):
                return self.existence.copy()
            
            def non_index_condition(self, obj, record_number, record):
                pass
            
            def get_record(self, recordset):
                return ()

        w = where.Where(query)
        w.lex()
        w.parse()
        self._create_clause_lookup(w)
        w.evaluate(Processors(self.trace, self.clauses))
        return w.node.get_clauses_from_root_in_walk_order()

    def _create_clause_lookup(self, obj):
        ce = self.clauses
        c = obj.node.get_clauses_from_root_in_walk_order()
        for e, i in enumerate(c):
            ce[hex(id(i))] = e

    def _cwc(self, where_clause, expected):
        ae = self.assertEqual
        wc = where.WhereClause()
        wc.__dict__.update(expected)
        ae(wc.operator, where_clause.operator)
        ae(wc.field, where_clause.field)
        ae(wc.condition, where_clause.condition)
        ae(wc.value, where_clause.value)
        ae(wc.not_phrase, where_clause.not_phrase)
        ae(wc.not_condition, where_clause.not_condition)
        ae(wc.not_value, where_clause.not_value)
        ae(wc.num, where_clause.num)
        ae(wc.alpha, where_clause.alpha)

    def _tec(self, expected):
        ae = self.assertEqual
        for e, t in enumerate(self.trace):
            ae(t, expected[e])


class Where_evaluateTC(unittest.TestCase):


    class Processors():

        file = {'f1', 'f2', 'f3', 'f4', 'f5', 'f6'}
        records = [
            {'f1':['b'],
             'f2':['v2'],
             'f3':['c', 'ck'],
             'f4':['d', 'dp'],
             'f5':['as']}, # record 0
            {'f1':['v1'],
             'f2':['d'],
             'f3':['v3'],
             'f4':['v4'],
             'f5':['as']}, # record 1
            {'f1':['v1'],
             'f3':['v3'],
             'f4':['v4']}, # record 2
            {'f1':['v1'],
             'f2':['v2'],
             'f4':['v4']}, # record 3
            {'f1':['v1'],
             'f2':['v2'],
             'f3':['v3']}, # record 4
            {'f1':['v1'],
             'f2':['v2'],
             'f3':['v3'],
             'f4':['v4'],
             'f5':['has']}, # record 5
            {'f1':['v1'],
             'f2':['v2'],
             'f3':['v3'],
             'f4':['v4'],
             'f5':['haste']}, # record 6
            {'f1':['v1'],
             'f2':['v2'],
             'f3':['v3'],
             'f4':['v4'],
             'f5':['hate']}, # record 7
            {'f1':['v1'],
             'f2':['v2'],
             'f3':['v3'],
             'f4':['v4'],
             'f5':['ate']}, # record 8
            {'f1':['v1'],
             'f2':['v2'],
             'f3':['v3'],
             'f4':['v4']}, # record 9
            ]
        existence = {e for e, r in enumerate(records)}

        def __init__(self):
            # True means Alpha test, False means Num, None means neither.
            # In particular f eq v defaults to f alpha eq v but f is v can be
            # seen as f eq v without the defaulting.
            # Alpha: b>a b>ab 2>1 2>12
            # Num:   b>a ab>b 2>1 12>2
            # In this test Num works for integer numbers only.
            self.compare_field_value = {
                (where.IS, None): self._is,
                (where.LIKE, None): self._like,
                (where.PRESENT, None): self._present,
                (where.EQ, True): self._eq,
                (where.NE, True): self._ne,
                (where.GT, True): self._gt,
                (where.LT, True): self._lt,
                (where.LE, True): self._le,
                (where.GE, True): self._ge,
                (where.BEFORE, True): self._before,
                (where.AFTER, True): self._after,
                ((where.FROM, where.TO), True): self._from_to,
                ((where.FROM, where.BELOW), True): self._from_below,
                ((where.ABOVE, where.TO), True): self._above_to,
                ((where.ABOVE, where.BELOW), True): self._above_below,
                (where.EQ, False): self._eq,
                (where.NE, False): self._ne,
                (where.GT, False): self._gt,
                (where.LT, False): self._lt,
                (where.LE, False): self._le,
                (where.GE, False): self._ge,
                (where.BEFORE, False): self._before,
                (where.AFTER, False): self._after,
                ((where.FROM, where.TO), False): self._from_to,
                ((where.FROM, where.BELOW), False): self._from_below,
                ((where.ABOVE, where.TO), False): self._above_to,
                ((where.ABOVE, where.BELOW), False): self._above_below,
                }
            self.boolean_operation = {
                where.AND: self._and,
                where.NOR: self._nor,
                where.OR: self._or,
                }
            
        def condition(self, obj):
            if obj.field not in self.file:
                return None
            if obj.condition in {where.IS, where.LIKE, where.PRESENT}:
                case = (obj.condition, None)
            elif obj.num is True:
                case = (obj.condition, False)
            else:
                case = (obj.condition, True)
            recordset = self.compare_field_value[case](obj)
            # Let parser stop 'field not is not value' 'field not is value'
            # 'field is not value' is only case of
            # 'field <condition> not value'
            if obj.not_condition:
                recordset = self.existence - recordset
            if obj.not_phrase:
                recordset = self.existence - recordset
            obj.result.answer = recordset
            
        def not_condition(self, obj):
            if obj.not_condition:
                obj.result.answer = self.existence - obj.result.answer
            if obj.not_phrase:
                obj.result.answer = self.existence - obj.result.answer
        
        def operator(self, obj):
            obj.left.result.answer = self.boolean_operation[obj.operator](obj)
            obj.result = obj.left.result
        
        def answer(self, obj):
            # Evaluate 'not' in clauses like '... and not (...) ...'
            if obj.not_phrase:
                obj.result.answer = self.existence - obj.result.answer
            obj.up.result = obj.result
        
        def initialize_answer(self, obj):
            obj.result.answer = set()
        
        def get_existence(self):
            return self.existence.copy()
        
        def get_record(self, recordset):
            for record_number in recordset:
                yield record_number, self.records[record_number]
            
        def non_index_condition(self, obj, record_number, record):
            if obj.condition == where.IS:
                if obj.not_value:
                    self._is_not(obj, record_number, record)
            elif obj.condition == where.NE:
                self._ne(obj, record_number, record)
            elif obj.condition == where.PRESENT:
                self._present(obj, record_number, record)
            elif obj.condition == where.LIKE:
                self._like(obj, record_number, record)

        def _is(self, obj):
            # 'field is value' and 'field is not value' are allowed
            recordset = set()
            if obj.not_value:
                raise RuntimeError('_is')
            else:
                for e, r in enumerate(self.records):
                    f = r.get(obj.field)
                    if f:
                        if obj.value in f:
                            recordset.add(e)
            return recordset

        def _is_not(self, obj, record_number, record):
            # 'field is value' and 'field is not value' are allowed
            if obj.not_value:
                f = record.get(obj.field)
                if f:
                    for v in f:
                        if v != obj.value:
                            obj.result.answer.add(record_number)
                            break
            else:
                raise RuntimeError('_is_not')

        def _like(self, obj, record_number, record):
            # Interpret 'like' as 'field value contains' for this test.
            f = record.get(obj.field)
            if f:
                for v in f:
                    if obj.value in v:
                        obj.result.answer.add(record_number)
                        break

        def _present(self, obj, record_number, record):
            if record.get(obj.field):
                obj.result.answer.add(record_number)

        def _eq(self, obj):
            recordset = set()
            for e, r in enumerate(self.records):
                f = r.get(obj.field)
                if f:
                    if obj.value in f:
                        recordset.add(e)
            return recordset

        def _ne(self, obj, record_number, record):
            f = record.get(obj.field)
            if f:
                for v in f:
                    if v != obj.value:
                        obj.result.answer.add(record_number)
                        break

        def _gt(self, obj):
            recordset = set()
            for e, r in enumerate(self.records):
                f = r.get(obj.field)
                if f:
                    for v in f:
                        if v > obj.value:
                            recordset.add(e)
                            break
            return recordset

        def _lt(self, obj):
            recordset = set()
            for e, r in enumerate(self.records):
                f = r.get(obj.field)
                if f:
                    for v in f:
                        if v < obj.value:
                            recordset.add(e)
                            break
            return recordset

        def _le(self, obj):
            recordset = set()
            for e, r in enumerate(self.records):
                f = r.get(obj.field)
                if f:
                    for v in f:
                        if v <= obj.value:
                            recordset.add(e)
                            break
            return recordset

        def _ge(self, obj):
            recordset = set()
            for e, r in enumerate(self.records):
                f = r.get(obj.field)
                if f:
                    for v in f:
                        if v >= obj.value:
                            recordset.add(e)
                            break
            return recordset

        def _before(self, obj):
            # Can use same logic as _lt in this test.
            recordset = set()
            for e, r in enumerate(self.records):
                f = r.get(obj.field)
                if f:
                    for v in f:
                        if v < obj.value:
                            recordset.add(e)
                            break
            return recordset

        def _after(self, obj):
            # Can use same logic as _gt in this test.
            recordset = set()
            for e, r in enumerate(self.records):
                f = r.get(obj.field)
                if f:
                    for v in f:
                        if v > obj.value:
                            recordset.add(e)
                            break
            return recordset

        def _from_to(self, obj):
            recordset = set()
            from_, to = obj.value
            for e, r in enumerate(self.records):
                f = r.get(obj.field)
                if f:
                    for v in f:
                        if v >= from_ and v <= to:
                            recordset.add(e)
                            break
            return recordset

        def _from_below(self, obj):
            recordset = set()
            from_, below = obj.value
            for e, r in enumerate(self.records):
                f = r.get(obj.field)
                if f:
                    for v in f:
                        if v >= from_ and v < below:
                            recordset.add(e)
                            break
            return recordset

        def _above_to(self, obj):
            recordset = set()
            above, to = obj.value
            for e, r in enumerate(self.records):
                f = r.get(obj.field)
                if f:
                    for v in f:
                        if v > above and v <= to:
                            recordset.add(e)
                            break
            return recordset

        def _above_below(self, obj):
            recordset = set()
            above, below = obj.value
            for e, r in enumerate(self.records):
                f = r.get(obj.field)
                if f:
                    for v in f:
                        if v > above and v < below:
                            recordset.add(e)
                            break
            return recordset

        def _and(self, obj):
            return obj.left.result.answer.intersection(obj.result.answer)

        def _nor(self, obj):
            return obj.left.result.answer.intersection(
                self.existence - obj.result.answer)

        def _or(self, obj):
            return obj.left.result.answer.union(obj.result.answer)


    class Constraints(Processors):
        # Override the actions to evaluate non-index conditions.
        # This allows the constraints on record scans to be preserved for
        # comparison with expected values.
        # The constraints refer to objects which are modified in real actions.
            
        def non_index_condition(self, obj, record_number, record):
            pass
            
        def not_condition(self, obj):
            pass
            
        def operator(self, obj):
            pass
            
        def answer(self, obj):
            pass


    def setUp(self):
            
        self.processors = Where_evaluateTC.Processors()
        self.constraints = Where_evaluateTC.Constraints()

    def tearDown(self):
        pass

    def test____assumptions(self):
        self.assertEqual(
            self.processors.file, {'f1', 'f2', 'f3', 'f4', 'f5', 'f6'})
        self.assertEqual(
            self.processors.records,
            [{'f1':['b'],
              'f2':['v2'],
              'f3':['c', 'ck'],
              'f4':['d', 'dp'],
              'f5':['as']}, # record 0
             {'f1':['v1'],
              'f2':['d'],
              'f3':['v3'],
              'f4':['v4'],
              'f5':['as']}, # record 1
             {'f1':['v1'],
              'f3':['v3'],
              'f4':['v4']}, # record 2
             {'f1':['v1'],
              'f2':['v2'],
              'f4':['v4']}, # record 3
             {'f1':['v1'],
              'f2':['v2'],
              'f3':['v3']}, # record 4
             {'f1':['v1'],
              'f2':['v2'],
              'f3':['v3'],
              'f4':['v4'],
              'f5':['has']}, # record 5
             {'f1':['v1'],
              'f2':['v2'],
              'f3':['v3'],
              'f4':['v4'],
              'f5':['haste']}, # record 6
             {'f1':['v1'],
              'f2':['v2'],
              'f3':['v3'],
              'f4':['v4'],
              'f5':['hate']}, # record 7
             {'f1':['v1'],
              'f2':['v2'],
              'f3':['v3'],
              'f4':['v4'],
              'f5':['ate']}, # record 8
             {'f1':['v1'],
              'f2':['v2'],
              'f3':['v3'],
              'f4':['v4']}, # record 9
             ])
        self.assertEqual(
            self.processors.compare_field_value,
            {(where.IS, None): self.processors._is,
             (where.LIKE, None): self.processors._like,
             (where.PRESENT, None): self.processors._present,
             (where.EQ, True): self.processors._eq,
             (where.NE, True): self.processors._ne,
             (where.GT, True): self.processors._gt,
             (where.LT, True): self.processors._lt,
             (where.LE, True): self.processors._le,
             (where.GE, True): self.processors._ge,
             (where.BEFORE, True): self.processors._before,
             (where.AFTER, True): self.processors._after,
             ((where.FROM, where.TO), True): self.processors._from_to,
             ((where.FROM, where.BELOW), True): self.processors._from_below,
             ((where.ABOVE, where.TO), True): self.processors._above_to,
             ((where.ABOVE, where.BELOW), True): self.processors._above_below,
             (where.EQ, False): self.processors._eq,
             (where.NE, False): self.processors._ne,
             (where.GT, False): self.processors._gt,
             (where.LT, False): self.processors._lt,
             (where.LE, False): self.processors._le,
             (where.GE, False): self.processors._ge,
             (where.BEFORE, False): self.processors._before,
             (where.AFTER, False): self.processors._after,
             ((where.FROM, where.TO), False): self.processors._from_to,
             ((where.FROM, where.BELOW), False): self.processors._from_below,
             ((where.ABOVE, where.TO), False): self.processors._above_to,
             ((where.ABOVE, where.BELOW), False): self.processors._above_below,
             })
        self.assertEqual(
            self.processors.boolean_operation,
            {where.AND: self.processors._and,
             where.NOR: self.processors._nor,
             where.OR: self.processors._or,
             })
        self.assertEqual(self.processors.existence, set(range(10)))
        self.assertEqual(self.processors.get_existence(), set(range(10)))

    def test_evaluate_01_simple(self):
        # See test_evaluate_01_simple_alpha and test_evaluate_01_simple_num
        self.assertEqual(self._evaluate('f1 is b'), set((0,)))
        self.assertEqual(self._evaluate('f1 is not b'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f1 eq b'), set((0,)))
        self.assertEqual(self._evaluate('f1 not eq b'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f2 is d'), set((1,)))
        self.assertEqual(self._evaluate('f2 is not d'),
                         set((0, 3, 4, 5, 6, 7, 8, 9))) # see 'f2 not eq d'
        self.assertEqual(self._evaluate('f2 eq d'), set((1,)))
        self.assertEqual(self._evaluate('f2 not eq d'),
                         set((0, 2, 3, 4, 5, 6, 7, 8, 9))) # see 'f2 is not d'
        self.assertEqual(self._evaluate('f3 after a'),
                         set((0, 1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 after c'),
                         set((0, 1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 after cg'),
                         set((0, 1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 after ck'),
                         set((1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 after v2'),
                         set((1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 after v3'),
                         set()) # see 'f3 not before v4' and similar negatives
        self.assertEqual(self._evaluate('f3 before a'), set())
        self.assertEqual(self._evaluate('f3 before c'), set())
        self.assertEqual(self._evaluate('f3 before cj'), set((0,)))
        self.assertEqual(self._evaluate('f3 before ck'), set((0,)))
        self.assertEqual(self._evaluate('f3 before v3'), set((0,)))
        self.assertEqual(self._evaluate('f3 before v4'),
                         set((0, 1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 not after a'),
                         set((3,)))
        self.assertEqual(self._evaluate('f3 not after c'),
                         set((3,)))
        self.assertEqual(self._evaluate('f3 not after cg'),
                         set((3,)))
        self.assertEqual(self._evaluate('f3 not after ck'),
                         set((0, 3,)))
        self.assertEqual(self._evaluate('f3 not after v2'),
                         set((0, 3,)))
        self.assertEqual(self._evaluate('f3 not after v3'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 not before a'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 not before c'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 not before cj'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 not before ck'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 not before v3'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 not before v4'),
                         set((3,))) # see 'f3 after v3' and similar positives
        self.assertEqual(self._evaluate('f4 from a to z'),
                         set((0, 1, 2, 3, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 from a to dp'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 from a to da'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 from a to d'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 from a to c'),
                         set())
        self.assertEqual(self._evaluate('f4 from a below z'),
                         set((0, 1, 2, 3, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 from a below dp'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 from a below da'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 from a below d'),
                         set())
        self.assertEqual(self._evaluate('f4 from a below c'),
                         set())
        self.assertEqual(self._evaluate('f4 above a to z'),
                         set((0, 1, 2, 3, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 above a to v'),
                         set((0,))) # use v rather than z as upper limit
        self.assertEqual(self._evaluate('f4 above d to v'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 above da to v'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 above dp to v'),
                         set())
        self.assertEqual(self._evaluate('f4 above e to v'),
                         set())
        self.assertEqual(self._evaluate('f4 above a below v'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 above d below v'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 above da below v'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 above dp below v'),
                         set())
        self.assertEqual(self._evaluate('f4 above e below v'),
                         set())
        self.assertEqual(self._evaluate('f4 not from a to z'),
                         set((4,)))
        self.assertEqual(self._evaluate('f4 not from a to dp'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not from a to da'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not from a to d'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not from a to c'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not from a below z'),
                         set((4,)))
        self.assertEqual(self._evaluate('f4 not from a below dp'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not from a below da'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not from a below d'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not from a below c'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not above a to z'),
                         set((4,))) # next use v rather than z as upper limit
        self.assertEqual(self._evaluate('f4 not above a to v'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not above d to v'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not above da to v'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not above dp to v'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not above e to v'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not above a below v'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not above d below v'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not above da below v'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not above dp below v'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not above e below v'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f5 ne as'),
                         set((5, 6, 7, 8)))
        self.assertEqual(self._evaluate('f5 not ne as'),
                         set((0, 1, 2, 3, 4, 9)))
        self.assertEqual(self._evaluate('f5 eq as'),
                         set((0, 1))) # 'eq' may not be equivalent to 'not ne'
        self.assertEqual(self._evaluate('f5 present'),
                         set((0, 1, 5, 6, 7, 8)))
        self.assertEqual(self._evaluate('f5 not present'),
                         set((2, 3, 4, 9)))
        self.assertEqual(self._evaluate('f5 like as'),
                         set((0, 1, 5, 6)))
        self.assertEqual(self._evaluate('f5 like ha'),
                         set((5, 6, 7)))
        self.assertEqual(self._evaluate('f5 not like as'),
                         set((2, 3, 4, 7, 8, 9)))

    def test_evaluate_01_simple_alpha(self):
        # Same as test_evaluate_01_simple but excludes 'is' 'like' 'present'
        self.assertEqual(self._evaluate('f1 alpha eq b'), set((0,)))
        self.assertEqual(self._evaluate('f1 not alpha eq b'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f2 alpha eq d'), set((1,)))
        self.assertEqual(self._evaluate('f2 not alpha eq d'),
                         set((0, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 alpha after a'),
                         set((0, 1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 alpha after c'),
                         set((0, 1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 alpha after cg'),
                         set((0, 1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 alpha after ck'),
                         set((1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 alpha after v2'),
                         set((1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 alpha after v3'),
                         set()) # see 'f3 not before v4' and similar negatives
        self.assertEqual(self._evaluate('f3 alpha before a'), set())
        self.assertEqual(self._evaluate('f3 alpha before c'), set())
        self.assertEqual(self._evaluate('f3 alpha before cj'), set((0,)))
        self.assertEqual(self._evaluate('f3 alpha before ck'), set((0,)))
        self.assertEqual(self._evaluate('f3 alpha before v3'), set((0,)))
        self.assertEqual(self._evaluate('f3 alpha before v4'),
                         set((0, 1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 not alpha after a'),
                         set((3,)))
        self.assertEqual(self._evaluate('f3 not alpha after c'),
                         set((3,)))
        self.assertEqual(self._evaluate('f3 not alpha after cg'),
                         set((3,)))
        self.assertEqual(self._evaluate('f3 not alpha after ck'),
                         set((0, 3,)))
        self.assertEqual(self._evaluate('f3 not alpha after v2'),
                         set((0, 3,)))
        self.assertEqual(self._evaluate('f3 not alpha after v3'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 not alpha before a'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 not alpha before c'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 not alpha before cj'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 not alpha before ck'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 not alpha before v3'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 not alpha before v4'),
                         set((3,))) # see 'f3 after v3' and similar positives
        self.assertEqual(self._evaluate('f4 alpha from a to z'),
                         set((0, 1, 2, 3, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 alpha from a to dp'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 alpha from a to da'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 alpha from a to d'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 alpha from a to c'),
                         set())
        self.assertEqual(self._evaluate('f4 alpha from a below z'),
                         set((0, 1, 2, 3, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 alpha from a below dp'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 alpha from a below da'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 alpha from a below d'),
                         set())
        self.assertEqual(self._evaluate('f4 alpha from a below c'),
                         set())
        self.assertEqual(self._evaluate('f4 alpha above a to z'),
                         set((0, 1, 2, 3, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 alpha above a to v'),
                         set((0,))) # use v rather than z as upper limit
        self.assertEqual(self._evaluate('f4 alpha above d to v'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 alpha above da to v'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 alpha above dp to v'),
                         set())
        self.assertEqual(self._evaluate('f4 alpha above e to v'),
                         set())
        self.assertEqual(self._evaluate('f4 alpha above a below v'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 alpha above d below v'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 alpha above da below v'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 alpha above dp below v'),
                         set())
        self.assertEqual(self._evaluate('f4 alpha above e below v'),
                         set())
        self.assertEqual(self._evaluate('f4 not alpha from a to z'),
                         set((4,)))
        self.assertEqual(self._evaluate('f4 not alpha from a to dp'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not alpha from a to da'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not alpha from a to d'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not alpha from a to c'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not alpha from a below z'),
                         set((4,)))
        self.assertEqual(self._evaluate('f4 not alpha from a below dp'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not alpha from a below da'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not alpha from a below d'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not alpha from a below c'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not alpha above a to z'),
                         set((4,))) # next use v rather than z as upper limit
        self.assertEqual(self._evaluate('f4 not alpha above a to v'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not alpha above d to v'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not alpha above da to v'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not alpha above dp to v'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not alpha above e to v'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not alpha above a below v'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not alpha above d below v'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not alpha above da below v'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not alpha above dp below v'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not alpha above e below v'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f5 alpha ne as'),
                         set((5, 6, 7, 8)))
        self.assertEqual(self._evaluate('f5 not alpha ne as'),
                         set((0, 1, 2, 3, 4, 9)))
        self.assertEqual(self._evaluate('f5 alpha eq as'),
                         set((0, 1))) # 'eq' may not be equivalent to 'not ne'

    def test_evaluate_01_simple_num(self):
        # Same as test_evaluate_01_simple but excludes 'is' 'like' 'present'
        self.assertEqual(self._evaluate('f1 num eq b'), set((0,)))
        self.assertEqual(self._evaluate('f1 not num eq b'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f2 num eq d'), set((1,)))
        self.assertEqual(self._evaluate('f2 not num eq d'),
                         set((0, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 num after a'),
                         set((0, 1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 num after c'),
                         set((0, 1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 num after cg'),
                         set((0, 1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 num after ck'),
                         set((1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 num after v2'),
                         set((1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 num after v3'),
                         set()) # see 'f3 not before v4' and similar negatives
        self.assertEqual(self._evaluate('f3 num before a'), set())
        self.assertEqual(self._evaluate('f3 num before c'), set())
        self.assertEqual(self._evaluate('f3 num before cj'), set((0,)))
        self.assertEqual(self._evaluate('f3 num before ck'), set((0,)))
        self.assertEqual(self._evaluate('f3 num before v3'), set((0,)))
        self.assertEqual(self._evaluate('f3 num before v4'),
                         set((0, 1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 not num after a'),
                         set((3,)))
        self.assertEqual(self._evaluate('f3 not num after c'),
                         set((3,)))
        self.assertEqual(self._evaluate('f3 not num after cg'),
                         set((3,)))
        self.assertEqual(self._evaluate('f3 not num after ck'),
                         set((0, 3,)))
        self.assertEqual(self._evaluate('f3 not num after v2'),
                         set((0, 3,)))
        self.assertEqual(self._evaluate('f3 not num after v3'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 not num before a'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 not num before c'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 not num before cj'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 not num before ck'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 not num before v3'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 not num before v4'),
                         set((3,))) # see 'f3 after v3' and similar positives
        self.assertEqual(self._evaluate('f4 num from a to z'),
                         set((0, 1, 2, 3, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 num from a to dp'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 num from a to da'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 num from a to d'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 num from a to c'),
                         set())
        self.assertEqual(self._evaluate('f4 num from a below z'),
                         set((0, 1, 2, 3, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 num from a below dp'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 num from a below da'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 num from a below d'),
                         set())
        self.assertEqual(self._evaluate('f4 num from a below c'),
                         set())
        self.assertEqual(self._evaluate('f4 num above a to z'),
                         set((0, 1, 2, 3, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 num above a to v'),
                         set((0,))) # use v rather than z as upper limit
        self.assertEqual(self._evaluate('f4 num above d to v'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 num above da to v'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 num above dp to v'),
                         set())
        self.assertEqual(self._evaluate('f4 num above e to v'),
                         set())
        self.assertEqual(self._evaluate('f4 num above a below v'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 num above d below v'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 num above da below v'),
                         set((0,)))
        self.assertEqual(self._evaluate('f4 num above dp below v'),
                         set())
        self.assertEqual(self._evaluate('f4 num above e below v'),
                         set())
        self.assertEqual(self._evaluate('f4 not num from a to z'),
                         set((4,)))
        self.assertEqual(self._evaluate('f4 not num from a to dp'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not num from a to da'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not num from a to d'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not num from a to c'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not num from a below z'),
                         set((4,)))
        self.assertEqual(self._evaluate('f4 not num from a below dp'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not num from a below da'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not num from a below d'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not num from a below c'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not num above a to z'),
                         set((4,))) # next use v rather than z as upper limit
        self.assertEqual(self._evaluate('f4 not num above a to v'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not num above d to v'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not num above da to v'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not num above dp to v'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not num above e to v'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not num above a below v'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not num above d below v'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not num above da below v'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not num above dp below v'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f4 not num above e below v'),
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f5 num ne as'),
                         set((5, 6, 7, 8)))
        self.assertEqual(self._evaluate('f5 not num ne as'),
                         set((0, 1, 2, 3, 4, 9)))
        self.assertEqual(self._evaluate('f5 num eq as'),
                         set((0, 1))) # 'eq' may not be equivalent to 'not ne'

    def test_evaluate_02_abbreviation(self):
        self.assertEqual(self._evaluate('f5 eq b or c'), set())
        self.assertEqual(self._evaluate('f3 eq c or v3'),
                         set((0, 1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('f3 eq c and ck'), set((0,)))
        self.assertEqual(self._evaluate('f3 eq c and v3'), set())

    def test_evaluate_03_nested_parentheses_01(self):
        # See test_evaluate_01_simple_alpha and test_evaluate_01_simple_num
        self.assertEqual(self._evaluate('(f1 is b)'), set((0,)))
        self.assertEqual(self._evaluate('(f1 is not b)'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('(f1 eq b)'), set((0,)))
        self.assertEqual(self._evaluate('(f1 not eq b)'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('(f2 is d)'), set((1,)))
        self.assertEqual(self._evaluate('(f2 is not d)'),
                         set((0, 3, 4, 5, 6, 7, 8, 9))) # see 'f2 not eq d'
        self.assertEqual(self._evaluate('(f2 eq d)'), set((1,)))
        self.assertEqual(self._evaluate('(f2 not eq d)'),
                         set((0, 2, 3, 4, 5, 6, 7, 8, 9))) # see 'f2 is not d'
        # Repeat with two ((
        self.assertEqual(self._evaluate('((f1 is b))'), set((0,)))
        self.assertEqual(self._evaluate('((f1 is not b))'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('((f1 eq b))'), set((0,)))
        self.assertEqual(self._evaluate('((f1 not eq b))'),
                         set((1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(self._evaluate('((f2 is d))'), set((1,)))
        self.assertEqual(self._evaluate('((f2 is not d))'),
                         set((0, 3, 4, 5, 6, 7, 8, 9))) # see 'f2 not eq d'
        self.assertEqual(self._evaluate('((f2 eq d))'), set((1,)))
        self.assertEqual(self._evaluate('((f2 not eq d))'),
                         set((0, 2, 3, 4, 5, 6, 7, 8, 9))) # see 'f2 is not d'
        # One direct comparison, expected answer in earlier test
        self.assertEqual(self._evaluate('(f4 not above d to v)'),
                         self._evaluate('((f4 not above d to v))'))

    def test_evaluate_04_constraint_01(self):
        w = self._evaluate_constraints('f5 eq as and f5 like at or f2 is d')
        cc = w.node.get_clauses_from_root_in_walk_order()
        self.assertEqual(cc[0].constraint.result.answer,
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[1].constraint.result.answer, set((0, 1)))
        self.assertEqual(cc[2].constraint.result.answer, set((0, 1)))
        self.assertEqual(cc[3].constraint.result.answer, set((1,)))
        self.assertEqual(len(cc), 4)
        self.assertEqual(
            (cc[0].operator, cc[0].field, cc[0].condition, cc[0].value),
            (None, None, None, None))
        self.assertEqual(
            (cc[1].operator, cc[1].field, cc[1].condition, cc[1].value),
            (None, 'f5', 'eq', 'as'))
        self.assertEqual(
            (cc[2].operator, cc[2].field, cc[2].condition, cc[2].value),
            ('and', 'f5', 'like', 'at'))
        self.assertEqual(
            (cc[3].operator, cc[3].field, cc[3].condition, cc[3].value),
            ('or', 'f2', 'is', 'd'))

    def test_evaluate_04_constraint_02(self):
        w = self._evaluate_constraints('f5 eq as and f5 like at and f2 is d')
        cc = w.node.get_clauses_from_root_in_walk_order()
        self.assertEqual(cc[0].constraint.result.answer,
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[1].constraint.result.answer, set((1,)))
        self.assertEqual(cc[2].constraint.result.answer, set((1,)))
        self.assertEqual(cc[3].constraint.result.answer, set((1,)))
        self.assertEqual(len(cc), 4)
        self.assertEqual(
            (cc[0].operator, cc[0].field, cc[0].condition, cc[0].value),
            (None, None, None, None))
        self.assertEqual(
            (cc[1].operator, cc[1].field, cc[1].condition, cc[1].value),
            (None, 'f5', 'eq', 'as'))
        self.assertEqual(
            (cc[2].operator, cc[2].field, cc[2].condition, cc[2].value),
            ('and', 'f2', 'is', 'd'))
        self.assertEqual(
            (cc[3].operator, cc[3].field, cc[3].condition, cc[3].value),
            ('and', 'f5', 'like', 'at'))

    def test_evaluate_04_constraint_03(self):
        w = self._evaluate_constraints('f5 like at and f5 eq as or f2 is d')
        cc = w.node.get_clauses_from_root_in_walk_order()
        self.assertEqual(cc[0].constraint.result.answer,
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[1].constraint.result.answer, set((0, 1)))
        self.assertEqual(cc[2].constraint.result.answer, set((0, 1)))
        self.assertEqual(cc[3].constraint.result.answer, set((1,)))
        self.assertEqual(len(cc), 4)
        self.assertEqual(
            (cc[0].operator, cc[0].field, cc[0].condition, cc[0].value),
            (None, None, None, None))
        self.assertEqual(
            (cc[1].operator, cc[1].field, cc[1].condition, cc[1].value),
            (None, 'f5', 'eq', 'as'))
        self.assertEqual(
            (cc[2].operator, cc[2].field, cc[2].condition, cc[2].value),
            ('and', 'f5', 'like', 'at'))
        self.assertEqual(
            (cc[3].operator, cc[3].field, cc[3].condition, cc[3].value),
            ('or', 'f2', 'is', 'd'))

    def test_evaluate_04_constraint_04(self):
        w = self._evaluate_constraints('f5 like at and f5 eq as and f2 is d')
        cc = w.node.get_clauses_from_root_in_walk_order()
        self.assertEqual(cc[0].constraint.result.answer,
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[1].constraint.result.answer, set((1,)))
        self.assertEqual(cc[2].constraint.result.answer, set((1,)))
        self.assertEqual(cc[3].constraint.result.answer, set((1,)))
        self.assertEqual(len(cc), 4)
        self.assertEqual(
            (cc[0].operator, cc[0].field, cc[0].condition, cc[0].value),
            (None, None, None, None))
        self.assertEqual(
            (cc[1].operator, cc[1].field, cc[1].condition, cc[1].value),
            (None, 'f2', 'is', 'd'))
        self.assertEqual(
            (cc[2].operator, cc[2].field, cc[2].condition, cc[2].value),
            ('and', 'f5', 'eq', 'as'))
        self.assertEqual(
            (cc[3].operator, cc[3].field, cc[3].condition, cc[3].value),
            ('and', 'f5', 'like', 'at'))

    def test_evaluate_04_constraint_05(self):
        w = self._evaluate_constraints('f5 eq as and f2 is d or f5 like at')
        cc = w.node.get_clauses_from_root_in_walk_order()
        self.assertEqual(cc[0].constraint.result.answer,
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[1].constraint.result.answer, set((1,)))
        self.assertEqual(cc[2].constraint.result.answer, set((1,)))
        self.assertEqual(cc[3].constraint.result.answer,
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(len(cc), 4)
        self.assertEqual(
            (cc[0].operator, cc[0].field, cc[0].condition, cc[0].value),
            (None, None, None, None))
        self.assertEqual(
            (cc[1].operator, cc[1].field, cc[1].condition, cc[1].value),
            (None, 'f5', 'eq', 'as'))
        self.assertEqual(
            (cc[2].operator, cc[2].field, cc[2].condition, cc[2].value),
            ('and', 'f2', 'is', 'd'))
        self.assertEqual(
            (cc[3].operator, cc[3].field, cc[3].condition, cc[3].value),
            ('or', 'f5', 'like', 'at'))

    def test_evaluate_04_constraint_06(self):
        w = self._evaluate_constraints('f5 eq as and f2 is d and f5 like at')
        cc = w.node.get_clauses_from_root_in_walk_order()
        self.assertEqual(cc[0].constraint.result.answer,
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[1].constraint.result.answer, set((1,)))
        self.assertEqual(cc[2].constraint.result.answer, set((1,)))
        self.assertEqual(cc[3].constraint.result.answer, set((1,)))
        self.assertEqual(len(cc), 4)
        self.assertEqual(
            (cc[0].operator, cc[0].field, cc[0].condition, cc[0].value),
            (None, None, None, None))
        self.assertEqual(
            (cc[1].operator, cc[1].field, cc[1].condition, cc[1].value),
            (None, 'f5', 'eq', 'as'))
        self.assertEqual(
            (cc[2].operator, cc[2].field, cc[2].condition, cc[2].value),
            ('and', 'f2', 'is', 'd'))
        self.assertEqual(
            (cc[3].operator, cc[3].field, cc[3].condition, cc[3].value),
            ('and', 'f5', 'like', 'at'))

    def test_evaluate_04_constraint_07(self):
        w = self._evaluate_constraints(
            'f3 eq c or v3 and ( f5 eq as and f5 like at or f2 is d )')
        cc = w.node.get_clauses_from_root_in_walk_order()
        self.assertEqual(cc[0].constraint.result.answer,
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[2].constraint.result.answer,
                         set((1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[3].constraint.result.answer,
                         set((1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[4].constraint.result.answer, set((0, 1)))
        self.assertEqual(cc[5].constraint.result.answer, set((0, 1)))
        self.assertEqual(cc[6].constraint.result.answer, set((1,)))
        self.assertEqual(len(cc), 7)
        self.assertEqual(
            (cc[0].operator, cc[0].field, cc[0].condition, cc[0].value),
            (None, None, None, None))
        self.assertEqual(
            (cc[1].operator, cc[1].field, cc[1].condition, cc[1].value),
            (None, 'f3', 'eq', 'c'))
        self.assertEqual(
            (cc[2].operator, cc[2].field, cc[2].condition, cc[2].value),
            ('or', 'f3', 'eq', 'v3'))
        self.assertEqual(
            (cc[3].operator, cc[3].field, cc[3].condition, cc[3].value),
            ('and', None, None, None))
        self.assertEqual(
            (cc[4].operator, cc[4].field, cc[4].condition, cc[4].value),
            (None, 'f5', 'eq', 'as'))
        self.assertEqual(
            (cc[5].operator, cc[5].field, cc[5].condition, cc[5].value),
            ('and', 'f5', 'like', 'at'))
        self.assertEqual(
            (cc[6].operator, cc[6].field, cc[6].condition, cc[6].value),
            ('or', 'f2', 'is', 'd'))

    def test_evaluate_04_constraint_08(self):
        w = self._evaluate_constraints(
            'f3 eq c or v3 and ( f5 eq as and f5 like at and f2 is d )')
        cc = w.node.get_clauses_from_root_in_walk_order()
        self.assertEqual(cc[0].constraint.result.answer,
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[1].constraint.result.answer, set((0,)))
        self.assertEqual(cc[2].constraint.result.answer,
                         set((1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[3].constraint.result.answer,
                         set((1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[4].constraint.result.answer, set((1,)))
        self.assertEqual(cc[5].constraint.result.answer, set((1,)))
        self.assertEqual(cc[6].constraint.result.answer, set((1,)))
        self.assertEqual(len(cc), 7)
        self.assertEqual(
            (cc[0].operator, cc[0].field, cc[0].condition, cc[0].value),
            (None, None, None, None))
        self.assertEqual(
            (cc[1].operator, cc[1].field, cc[1].condition, cc[1].value),
            (None, 'f3', 'eq', 'c'))
        self.assertEqual(
            (cc[2].operator, cc[2].field, cc[2].condition, cc[2].value),
            ('or', 'f3', 'eq', 'v3'))
        self.assertEqual(
            (cc[3].operator, cc[3].field, cc[3].condition, cc[3].value),
            ('and', None, None, None))
        self.assertEqual(
            (cc[4].operator, cc[4].field, cc[4].condition, cc[4].value),
            (None, 'f5', 'eq', 'as'))
        self.assertEqual(
            (cc[5].operator, cc[5].field, cc[5].condition, cc[5].value),
            ('and', 'f2', 'is', 'd'))
        self.assertEqual(
            (cc[6].operator, cc[6].field, cc[6].condition, cc[6].value),
            ('and', 'f5', 'like', 'at'))

    def test_evaluate_04_constraint_09(self):
        w = self._evaluate_constraints(
            'f3 eq c or v3 and ( f5 like at and f5 eq as or f2 is d )')
        cc = w.node.get_clauses_from_root_in_walk_order()
        self.assertEqual(cc[0].constraint.result.answer,
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[1].constraint.result.answer, set((0,)))
        self.assertEqual(cc[2].constraint.result.answer,
                         set((1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[3].constraint.result.answer,
                         set((1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[4].constraint.result.answer, set((0, 1)))
        self.assertEqual(cc[5].constraint.result.answer, set((0, 1)))
        self.assertEqual(cc[6].constraint.result.answer, set((1,)))
        self.assertEqual(len(cc), 7)
        self.assertEqual(
            (cc[0].operator, cc[0].field, cc[0].condition, cc[0].value),
            (None, None, None, None))
        self.assertEqual(
            (cc[1].operator, cc[1].field, cc[1].condition, cc[1].value),
            (None, 'f3', 'eq', 'c'))
        self.assertEqual(
            (cc[2].operator, cc[2].field, cc[2].condition, cc[2].value),
            ('or', 'f3', 'eq', 'v3'))
        self.assertEqual(
            (cc[3].operator, cc[3].field, cc[3].condition, cc[3].value),
            ('and', None, None, None))
        self.assertEqual(
            (cc[4].operator, cc[4].field, cc[4].condition, cc[4].value),
            (None, 'f5', 'eq', 'as'))
        self.assertEqual(
            (cc[5].operator, cc[5].field, cc[5].condition, cc[5].value),
            ('and', 'f5', 'like', 'at'))
        self.assertEqual(
            (cc[6].operator, cc[6].field, cc[6].condition, cc[6].value),
            ('or', 'f2', 'is', 'd'))

    def test_evaluate_04_constraint_10(self):
        w = self._evaluate_constraints(
            'f3 eq c or v3 and ( f5 like at and f5 eq as and f2 is d )')
        cc = w.node.get_clauses_from_root_in_walk_order()
        self.assertEqual(cc[0].constraint.result.answer,
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[1].constraint.result.answer, set((0,)))
        self.assertEqual(cc[2].constraint.result.answer,
                         set((1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[3].constraint.result.answer,
                         set((1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[4].constraint.result.answer, set((1,)))
        self.assertEqual(cc[5].constraint.result.answer, set((1,)))
        self.assertEqual(cc[6].constraint.result.answer, set((1,)))
        self.assertEqual(len(cc), 7)
        self.assertEqual(
            (cc[0].operator, cc[0].field, cc[0].condition, cc[0].value),
            (None, None, None, None))
        self.assertEqual(
            (cc[1].operator, cc[1].field, cc[1].condition, cc[1].value),
            (None, 'f3', 'eq', 'c'))
        self.assertEqual(
            (cc[2].operator, cc[2].field, cc[2].condition, cc[2].value),
            ('or', 'f3', 'eq', 'v3'))
        self.assertEqual(
            (cc[3].operator, cc[3].field, cc[3].condition, cc[3].value),
            ('and', None, None, None))
        self.assertEqual(
            (cc[4].operator, cc[4].field, cc[4].condition, cc[4].value),
            (None, 'f2', 'is', 'd'))
        self.assertEqual(
            (cc[5].operator, cc[5].field, cc[5].condition, cc[5].value),
            ('and', 'f5', 'eq', 'as'))
        self.assertEqual(
            (cc[6].operator, cc[6].field, cc[6].condition, cc[6].value),
            ('and', 'f5', 'like', 'at'))

    def test_evaluate_04_constraint_11(self):
        w = self._evaluate_constraints(
            'f3 eq c or v3 and ( f5 eq as and f2 is d or f5 like at )')
        cc = w.node.get_clauses_from_root_in_walk_order()
        self.assertEqual(cc[0].constraint.result.answer,
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[1].constraint.result.answer, set((0,)))
        self.assertEqual(cc[2].constraint.result.answer,
                         set((1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[3].constraint.result.answer,
                         set((1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[4].constraint.result.answer, set((1,)))
        self.assertEqual(cc[5].constraint.result.answer, set((1,)))
        self.assertEqual(cc[6].constraint.result.answer,
                         set((1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(len(cc), 7)
        self.assertEqual(
            (cc[0].operator, cc[0].field, cc[0].condition, cc[0].value),
            (None, None, None, None))
        self.assertEqual(
            (cc[1].operator, cc[1].field, cc[1].condition, cc[1].value),
            (None, 'f3', 'eq', 'c'))
        self.assertEqual(
            (cc[2].operator, cc[2].field, cc[2].condition, cc[2].value),
            ('or', 'f3', 'eq', 'v3'))
        self.assertEqual(
            (cc[3].operator, cc[3].field, cc[3].condition, cc[3].value),
            ('and', None, None, None))
        self.assertEqual(
            (cc[4].operator, cc[4].field, cc[4].condition, cc[4].value),
            (None, 'f5', 'eq', 'as'))
        self.assertEqual(
            (cc[5].operator, cc[5].field, cc[5].condition, cc[5].value),
            ('and', 'f2', 'is', 'd'))
        self.assertEqual(
            (cc[6].operator, cc[6].field, cc[6].condition, cc[6].value),
            ('or', 'f5', 'like', 'at'))

    def test_evaluate_04_constraint_12(self):
        w = self._evaluate_constraints(
            'f3 eq c or v3 and ( f5 eq as and f2 is d and f5 like at )')
        cc = w.node.get_clauses_from_root_in_walk_order()
        self.assertEqual(cc[0].constraint.result.answer,
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[1].constraint.result.answer, set((0,)))
        self.assertEqual(cc[2].constraint.result.answer,
                         set((1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[3].constraint.result.answer,
                         set((1, 2, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[4].constraint.result.answer, set((1,)))
        self.assertEqual(cc[5].constraint.result.answer, set((1,)))
        self.assertEqual(cc[6].constraint.result.answer, set((1,)))
        self.assertEqual(len(cc), 7)
        self.assertEqual(
            (cc[0].operator, cc[0].field, cc[0].condition, cc[0].value),
            (None, None, None, None))
        self.assertEqual(
            (cc[1].operator, cc[1].field, cc[1].condition, cc[1].value),
            (None, 'f3', 'eq', 'c'))
        self.assertEqual(
            (cc[2].operator, cc[2].field, cc[2].condition, cc[2].value),
            ('or', 'f3', 'eq', 'v3'))
        self.assertEqual(
            (cc[3].operator, cc[3].field, cc[3].condition, cc[3].value),
            ('and', None, None, None))
        self.assertEqual(
            (cc[4].operator, cc[4].field, cc[4].condition, cc[4].value),
            (None, 'f5', 'eq', 'as'))
        self.assertEqual(
            (cc[5].operator, cc[5].field, cc[5].condition, cc[5].value),
            ('and', 'f2', 'is', 'd'))
        self.assertEqual(
            (cc[6].operator, cc[6].field, cc[6].condition, cc[6].value),
            ('and', 'f5', 'like', 'at'))

    def test_evaluate_04_constraint_13(self):
        w = self._evaluate_constraints('f5 eq as and (f5 like at or f2 is d)')
        cc = w.node.get_clauses_from_root_in_walk_order()
        self.assertEqual(cc[0].constraint.result.answer,
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[1].constraint.result.answer, set((0, 1)))
        self.assertEqual(cc[2].constraint.result.answer, set((0, 1)))
        self.assertEqual(cc[3].constraint.result.answer, set((0, 1)))
        self.assertEqual(cc[4].constraint.result.answer, set((1,)))
        self.assertEqual(len(cc), 5)
        self.assertEqual(
            (cc[0].operator, cc[0].field, cc[0].condition, cc[0].value),
            (None, None, None, None))
        self.assertEqual((id(cc[0].left), id(cc[0].up)), (id(None), id(None)))
        self.assertEqual(
            (cc[1].operator, cc[1].field, cc[1].condition, cc[1].value),
            (None, 'f5', 'eq', 'as'))
        self.assertEqual((id(cc[1].left), id(cc[1].up)), (id(None), id(cc[0])))
        self.assertEqual(
            (cc[2].operator, cc[2].field, cc[2].condition, cc[2].value),
            ('and', None, None, None))
        self.assertEqual((id(cc[2].left), id(cc[2].up)), (id(cc[1]), id(cc[0])))
        self.assertEqual(
            (cc[3].operator, cc[3].field, cc[3].condition, cc[3].value),
            (None, 'f5', 'like', 'at'))
        self.assertEqual((id(cc[3].left), id(cc[3].up)), (id(None), id(cc[2])))
        self.assertEqual(
            (cc[4].operator, cc[4].field, cc[4].condition, cc[4].value),
            ('or', 'f2', 'is', 'd'))
        self.assertEqual((id(cc[4].left), id(cc[4].up)), (id(cc[3]), id(cc[2])))

    def test_evaluate_04_constraint_14(self):
        w = self._evaluate_constraints('f5 eq as and f2 is d or f5 like at')
        cc = w.node.get_clauses_from_root_in_walk_order()
        self.assertEqual(cc[0].constraint.result.answer,
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[1].constraint.result.answer, set((1,)))
        self.assertEqual(cc[2].constraint.result.answer, set((1,)))
        self.assertEqual(cc[3].constraint.result.answer,
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(len(cc), 4)

    def test_evaluate_04_constraint_15(self):
        w = self._evaluate_constraints('f5 eq as and (f2 is d or f5 like at)')
        cc = w.node.get_clauses_from_root_in_walk_order()
        self.assertEqual(cc[0].constraint.result.answer,
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[1].constraint.result.answer, set((0, 1)))
        self.assertEqual(cc[2].constraint.result.answer, set((0, 1)))
        self.assertEqual(cc[3].constraint.result.answer, set((1,)))
        self.assertEqual(cc[4].constraint.result.answer, set((0, 1)))
        self.assertEqual(len(cc), 5)

    def test_evaluate_04_constraint_16(self):
        w = self._evaluate_constraints('f5 eq as and (f2 is d and f5 like at)')
        cc = w.node.get_clauses_from_root_in_walk_order()
        self.assertEqual(cc[0].constraint.result.answer,
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[1].constraint.result.answer, set((0, 1)))
        self.assertEqual(cc[2].constraint.result.answer, set((0, 1)))
        self.assertEqual(cc[3].constraint.result.answer, set((1,)))
        self.assertEqual(cc[4].constraint.result.answer, set((1,)))
        self.assertEqual(len(cc), 5)
        self.assertEqual(
            (cc[0].operator, cc[0].field, cc[0].condition, cc[0].value),
            (None, None, None, None))
        self.assertEqual((id(cc[0].left), id(cc[0].up)), (id(None), id(None)))
        self.assertEqual(
            (cc[1].operator, cc[1].field, cc[1].condition, cc[1].value),
            (None, 'f5', 'eq', 'as'))
        self.assertEqual((id(cc[1].left), id(cc[1].up)), (id(None), id(cc[0])))
        self.assertEqual(
            (cc[2].operator, cc[2].field, cc[2].condition, cc[2].value),
            ('and', None, None, None))
        self.assertEqual((id(cc[2].left), id(cc[2].up)), (id(cc[1]), id(cc[0])))
        self.assertEqual(
            (cc[3].operator, cc[3].field, cc[3].condition, cc[3].value),
            (None, 'f2', 'is', 'd'))
        self.assertEqual((id(cc[3].left), id(cc[3].up)), (id(None), id(cc[2])))
        self.assertEqual(
            (cc[4].operator, cc[4].field, cc[4].condition, cc[4].value),
            ('and', 'f5', 'like', 'at'))
        self.assertEqual((id(cc[4].left), id(cc[4].up)), (id(cc[3]), id(cc[2])))

    def test_evaluate_04_constraint_17(self):
        w = self._evaluate_constraints('f4 eq d and (f2 is d and f5 like at)')
        cc = w.node.get_clauses_from_root_in_walk_order()
        self.assertEqual(cc[0].constraint.result.answer,
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[1].constraint.result.answer, set((0,)))
        self.assertEqual(cc[2].constraint.result.answer, set((0,)))
        self.assertEqual(cc[3].constraint.result.answer, set((1,)))
        self.assertEqual(cc[4].constraint.result.answer, set((1,)))
        self.assertEqual(len(cc), 5)
        self.assertEqual(
            (cc[0].operator, cc[0].field, cc[0].condition, cc[0].value),
            (None, None, None, None))
        self.assertEqual((id(cc[0].left), id(cc[0].up)), (id(None), id(None)))
        self.assertEqual(
            (cc[1].operator, cc[1].field, cc[1].condition, cc[1].value),
            (None, 'f4', 'eq', 'd'))
        self.assertEqual((id(cc[1].left), id(cc[1].up)), (id(None), id(cc[0])))
        self.assertEqual(
            (cc[2].operator, cc[2].field, cc[2].condition, cc[2].value),
            ('and', None, None, None))
        self.assertEqual((id(cc[2].left), id(cc[2].up)), (id(cc[1]), id(cc[0])))
        self.assertEqual(
            (cc[3].operator, cc[3].field, cc[3].condition, cc[3].value),
            (None, 'f2', 'is', 'd'))
        self.assertEqual((id(cc[3].left), id(cc[3].up)), (id(None), id(cc[2])))
        self.assertEqual(
            (cc[4].operator, cc[4].field, cc[4].condition, cc[4].value),
            ('and', 'f5', 'like', 'at'))
        self.assertEqual((id(cc[4].left), id(cc[4].up)), (id(cc[3]), id(cc[2])))

    def test_evaluate_04_constraint_18(self):
        w = self._evaluate_constraints(
            'f4 eq d and f2 is d and f5 like at or f5 eq has and f3 eq v3')
        cc = w.node.get_clauses_from_root_in_walk_order()
        self.assertEqual(cc[0].constraint.result.answer,
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[1].constraint.result.answer, set())
        self.assertEqual(cc[2].constraint.result.answer, set())
        self.assertEqual(cc[3].constraint.result.answer, set())
        self.assertEqual(cc[4].constraint.result.answer, set((5,)))
        self.assertEqual(cc[5].constraint.result.answer, set((5,)))
        self.assertEqual(len(cc), 6)
        self.assertEqual(
            (cc[0].operator, cc[0].field, cc[0].condition, cc[0].value),
            (None, None, None, None))
        self.assertEqual((id(cc[0].left), id(cc[0].up)), (id(None), id(None)))
        self.assertEqual(
            (cc[1].operator, cc[1].field, cc[1].condition, cc[1].value),
            (None, 'f4', 'eq', 'd'))
        self.assertEqual(
            (cc[2].operator, cc[2].field, cc[2].condition, cc[2].value),
            ('and', 'f2', 'is', 'd'))
        self.assertEqual(
            (cc[3].operator, cc[3].field, cc[3].condition, cc[3].value),
            ('and', 'f5', 'like', 'at'))
        self.assertEqual(
            (cc[4].operator, cc[4].field, cc[4].condition, cc[4].value),
            ('or', 'f5', 'eq', 'has'))
        self.assertEqual(
            (cc[5].operator, cc[5].field, cc[5].condition, cc[5].value),
            ('and', 'f3', 'eq', 'v3'))

    def test_evaluate_04_constraint_19(self):
        w = self._evaluate_constraints(
            'f4 eq d and f5 like at and f2 is d or f5 eq has and f3 eq v3')
        cc = w.node.get_clauses_from_root_in_walk_order()
        self.assertEqual(cc[0].constraint.result.answer,
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[1].constraint.result.answer, set())
        self.assertEqual(cc[2].constraint.result.answer, set())
        self.assertEqual(cc[3].constraint.result.answer, set())
        self.assertEqual(cc[4].constraint.result.answer, set((5,)))
        self.assertEqual(cc[5].constraint.result.answer, set((5,)))
        self.assertEqual(len(cc), 6)
        self.assertEqual(
            (cc[0].operator, cc[0].field, cc[0].condition, cc[0].value),
            (None, None, None, None))
        self.assertEqual((id(cc[0].left), id(cc[0].up)), (id(None), id(None)))
        self.assertEqual(
            (cc[1].operator, cc[1].field, cc[1].condition, cc[1].value),
            (None, 'f4', 'eq', 'd'))
        self.assertEqual(
            (cc[2].operator, cc[2].field, cc[2].condition, cc[2].value),
            ('and', 'f2', 'is', 'd'))
        self.assertEqual(
            (cc[3].operator, cc[3].field, cc[3].condition, cc[3].value),
            ('and', 'f5', 'like', 'at'))
        self.assertEqual(
            (cc[4].operator, cc[4].field, cc[4].condition, cc[4].value),
            ('or', 'f5', 'eq', 'has'))
        self.assertEqual(
            (cc[5].operator, cc[5].field, cc[5].condition, cc[5].value),
            ('and', 'f3', 'eq', 'v3'))

    def test_evaluate_04_constraint_20(self):
        w = self._evaluate_constraints(
            'f5 like at and f4 eq d and f2 is d or f5 eq has and f3 eq v3')
        cc = w.node.get_clauses_from_root_in_walk_order()
        self.assertEqual(cc[0].constraint.result.answer,
                         set((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)))
        self.assertEqual(cc[1].constraint.result.answer, set())
        self.assertEqual(cc[2].constraint.result.answer, set())
        self.assertEqual(cc[3].constraint.result.answer, set())
        self.assertEqual(cc[4].constraint.result.answer, set((5,)))
        self.assertEqual(cc[5].constraint.result.answer, set((5,)))
        self.assertEqual(len(cc), 6)
        self.assertEqual(
            (cc[0].operator, cc[0].field, cc[0].condition, cc[0].value),
            (None, None, None, None))
        self.assertEqual((id(cc[0].left), id(cc[0].up)), (id(None), id(None)))
        self.assertEqual(
            (cc[1].operator, cc[1].field, cc[1].condition, cc[1].value),
            (None, 'f2', 'is', 'd'))
        self.assertEqual(
            (cc[2].operator, cc[2].field, cc[2].condition, cc[2].value),
            ('and', 'f4', 'eq', 'd'))
        self.assertEqual(
            (cc[3].operator, cc[3].field, cc[3].condition, cc[3].value),
            ('and', 'f5', 'like', 'at'))
        self.assertEqual(
            (cc[4].operator, cc[4].field, cc[4].condition, cc[4].value),
            ('or', 'f5', 'eq', 'has'))
        self.assertEqual(
            (cc[5].operator, cc[5].field, cc[5].condition, cc[5].value),
            ('and', 'f3', 'eq', 'v3'))

    def test_evaluate_05_constraints(self):
        self.assertEqual(
            self._evaluate('f5 eq as and f5 like at or f2 is d'), set((1,)))
        self.assertEqual(
            self._evaluate('f5 eq as and f5 like at and f2 is d'), set())
        self.assertEqual(
            self._evaluate('f5 like at and f5 eq as or f2 is d'),
            set((1,)))
        self.assertEqual(
            self._evaluate('f5 like at and f5 eq as and f2 is d'), set())
        self.assertEqual(
            self._evaluate('f5 eq as and f2 is d or f5 like at'),
            set((1, 7, 8)))
        self.assertEqual(
            self._evaluate('f5 eq as and f2 is d and f5 like at'), set())
        self.assertEqual(
            self._evaluate(
                'f3 eq c or v3 and ( f5 eq as and f5 like at or f2 is d )'),
            set((1,)))
        self.assertEqual(
            self._evaluate(
                'f3 eq c or v3 and ( f5 eq as and f5 like at and f2 is d )'),
            set())
        self.assertEqual(
            self._evaluate(
                'f3 eq c or v3 and ( f5 like at and f5 eq as or f2 is d )'),
            set((1,)))
        self.assertEqual(
            self._evaluate(
                'f3 eq c or v3 and ( f5 like at and f5 eq as and f2 is d )'),
            set())
        self.assertEqual(
            self._evaluate(
                'f3 eq c or v3 and ( f5 eq as and f2 is d or f5 like at )'),
            set((1, 7, 8)))
        self.assertEqual(
            self._evaluate(
                'f3 eq c or v3 and ( f5 eq as and f2 is d and f5 like at )'),
            set())
        self.assertEqual(
            self._evaluate('f5 eq as and (f5 like at or f2 is d)'), set((1,)))
        self.assertEqual(
            self._evaluate('f5 eq as and f2 is d or f5 like at'),
            set((1, 7, 8)))
        self.assertEqual(
            self._evaluate('f5 eq as and (f2 is d or f5 like at)'), set((1,)))
        self.assertEqual(
            self._evaluate('f5 eq as and (f2 is d and f5 like at)'), set())
        self.assertEqual(
            self._evaluate('f4 eq d and (f2 is d and f5 like at)'), set())
        self.assertEqual(
            self._evaluate(
                'f4 eq d and f2 is d and f5 like at or f5 eq has and f3 eq v3'),
            set((5,)))
        self.assertEqual(
            self._evaluate(
                'f4 eq d and f5 like at and f2 is d or f5 eq has and f3 eq v3'),
            set((5,)))
        self.assertEqual(
            self._evaluate(
                'f5 like at and f4 eq d and f2 is d or f5 eq has and f3 eq v3'),
            set((5,)))

    def _evaluate(self, query):
        w = where.Where(query)
        w.lex()
        w.parse()
        w.evaluate(self.processors)
        return w.node.result.answer

    def _evaluate_constraints(self, query):
        w = where.Where(query)
        w.lex()
        w.parse()
        w.evaluate(self.constraints)
        return w


class Where_validateTC(unittest.TestCase):


    class Processors():

        file = {'f1', 'f2', 'f3', 'f4', 'f5', 'f6'}
        records = [
            {'f1':['b'],
             'f2':['v2'],
             'f3':['c', 'ck'],
             'f4':['d', 'dp'],
             'f5':['as']}, # record 0
            {'f1':['v1'],
             'f2':['d'],
             'f3':['v3'],
             'f4':['v4'],
             'f5':['as']}, # record 1
            {'f1':['v1'],
             'f3':['v3'],
             'f4':['v4']}, # record 2
            {'f1':['v1'],
             'f2':['v2'],
             'f4':['v4']}, # record 3
            {'f1':['v1'],
             'f2':['v2'],
             'f3':['v3']}, # record 4
            {'f1':['v1'],
             'f2':['v2'],
             'f3':['v3'],
             'f4':['v4'],
             'f5':['has']}, # record 5
            {'f1':['v1'],
             'f2':['v2'],
             'f3':['v3'],
             'f4':['v4'],
             'f5':['haste']}, # record 6
            {'f1':['v1'],
             'f2':['v2'],
             'f3':['v3'],
             'f4':['v4'],
             'f5':['hate']}, # record 7
            {'f1':['v1'],
             'f2':['v2'],
             'f3':['v3'],
             'f4':['v4'],
             'f5':['ate']}, # record 8
            {'f1':['v1'],
             'f2':['v2'],
             'f3':['v3'],
             'f4':['v4']}, # record 9
            ]
        existence = {e for e, r in enumerate(records)}

        def __init__(self):
            # True means Alpha test, False means Num, None means neither.
            # In particular f eq v defaults to f alpha eq v but f is v can be
            # seen as f eq v without the defaulting.
            # Alpha: b>a b>ab 2>1 2>12
            # Num:   b>a ab>b 2>1 12>2
            # In this test Num works for integer numbers only.
            self.compare_field_value = {
                (where.IS, None): self._is,
                (where.LIKE, None): self._like,
                (where.PRESENT, None): self._present,
                (where.EQ, True): self._eq,
                (where.NE, True): self._ne,
                (where.GT, True): self._gt,
                (where.LT, True): self._lt,
                (where.LE, True): self._le,
                (where.GE, True): self._ge,
                (where.BEFORE, True): self._before,
                (where.AFTER, True): self._after,
                ((where.FROM, where.TO), True): self._from_to,
                ((where.FROM, where.BELOW), True): self._from_below,
                ((where.ABOVE, where.TO), True): self._above_to,
                ((where.ABOVE, where.BELOW), True): self._above_below,
                (where.EQ, False): self._eq,
                (where.NE, False): self._ne,
                (where.GT, False): self._gt,
                (where.LT, False): self._lt,
                (where.LE, False): self._le,
                (where.GE, False): self._ge,
                (where.BEFORE, False): self._before,
                (where.AFTER, False): self._after,
                ((where.FROM, where.TO), False): self._from_to,
                ((where.FROM, where.BELOW), False): self._from_below,
                ((where.ABOVE, where.TO), False): self._above_to,
                ((where.ABOVE, where.BELOW), False): self._above_below,
                }
            self.boolean_operation = {
                where.AND: self._and,
                where.NOR: self._nor,
                where.OR: self._or,
                }
            
        def condition(self, obj):
            if obj.field not in self.file:
                return None
            if obj.condition in {where.IS, where.LIKE, where.PRESENT}:
                case = (obj.condition, None)
            elif obj.num is True:
                case = (obj.condition, False)
            else:
                case = (obj.condition, True)
            recordset = self.compare_field_value[case](obj)
            # Let parser stop 'field not is not value' 'field not is value'
            # 'field is not value' is only case of
            # 'field <condition> not value'
            if obj.not_condition:
                recordset = self.existence - recordset
            if obj.not_phrase:
                recordset = self.existence - recordset
            obj.result.answer = recordset
            
        def not_condition(self, obj):
            if obj.not_condition:
                obj.result.answer = self.existence - obj.result.answer
            if obj.not_phrase:
                obj.result.answer = self.existence - obj.result.answer
        
        def operator(self, obj):
            obj.left.result.answer = self.boolean_operation[obj.operator](obj)
            obj.result = obj.left.result
        
        def answer(self, obj):
            # Evaluate 'not' in clauses like '... and not (...) ...'
            if obj.not_phrase:
                obj.result.answer = self.existence - obj.result.answer
            obj.up.result = obj.result
        
        def initialize_answer(self, obj):
            obj.result.answer = set()
        
        def get_existence(self):
            return self.existence.copy()
        
        def get_record(self, recordset):
            for record_number in recordset:
                yield record_number, self.records[record_number]
            
        def non_index_condition(self, obj, record_number, record):
            if obj.condition == where.IS:
                if obj.not_value:
                    self._is_not(obj, record_number, record)
            elif obj.condition == where.NE:
                self._ne(obj, record_number, record)
            elif obj.condition == where.PRESENT:
                self._present(obj, record_number, record)
            elif obj.condition == where.LIKE:
                self._like(obj, record_number, record)

        def _is(self, obj):
            # 'field is value' and 'field is not value' are allowed
            recordset = set()
            if obj.not_value:
                raise RuntimeError('_is')
            else:
                for e, r in enumerate(self.records):
                    f = r.get(obj.field)
                    if f:
                        if obj.value in f:
                            recordset.add(e)
            return recordset

        def _is_not(self, obj, record_number, record):
            # 'field is value' and 'field is not value' are allowed
            if obj.not_value:
                f = record.get(obj.field)
                if f:
                    for v in f:
                        if v != obj.value:
                            obj.result.answer.add(record_number)
                            break
            else:
                raise RuntimeError('_is_not')

        def _like(self, obj, record_number, record):
            # Interpret 'like' as 'field value contains' for this test.
            f = record.get(obj.field)
            if f:
                for v in f:
                    if obj.value in v:
                        obj.result.answer.add(record_number)
                        break

        def _present(self, obj, record_number, record):
            if record.get(obj.field):
                obj.result.answer.add(record_number)

        def _eq(self, obj):
            recordset = set()
            for e, r in enumerate(self.records):
                f = r.get(obj.field)
                if f:
                    if obj.value in f:
                        recordset.add(e)
            return recordset

        def _ne(self, obj, record_number, record):
            f = record.get(obj.field)
            if f:
                for v in f:
                    if v != obj.value:
                        obj.result.answer.add(record_number)
                        break

        def _gt(self, obj):
            recordset = set()
            for e, r in enumerate(self.records):
                f = r.get(obj.field)
                if f:
                    for v in f:
                        if v > obj.value:
                            recordset.add(e)
                            break
            return recordset

        def _lt(self, obj):
            recordset = set()
            for e, r in enumerate(self.records):
                f = r.get(obj.field)
                if f:
                    for v in f:
                        if v < obj.value:
                            recordset.add(e)
                            break
            return recordset

        def _le(self, obj):
            recordset = set()
            for e, r in enumerate(self.records):
                f = r.get(obj.field)
                if f:
                    for v in f:
                        if v <= obj.value:
                            recordset.add(e)
                            break
            return recordset

        def _ge(self, obj):
            recordset = set()
            for e, r in enumerate(self.records):
                f = r.get(obj.field)
                if f:
                    for v in f:
                        if v >= obj.value:
                            recordset.add(e)
                            break
            return recordset

        def _before(self, obj):
            # Can use same logic as _lt in this test.
            recordset = set()
            for e, r in enumerate(self.records):
                f = r.get(obj.field)
                if f:
                    for v in f:
                        if v < obj.value:
                            recordset.add(e)
                            break
            return recordset

        def _after(self, obj):
            # Can use same logic as _gt in this test.
            recordset = set()
            for e, r in enumerate(self.records):
                f = r.get(obj.field)
                if f:
                    for v in f:
                        if v > obj.value:
                            recordset.add(e)
                            break
            return recordset

        def _from_to(self, obj):
            recordset = set()
            from_, to = obj.value
            for e, r in enumerate(self.records):
                f = r.get(obj.field)
                if f:
                    for v in f:
                        if v >= from_ and v <= to:
                            recordset.add(e)
                            break
            return recordset

        def _from_below(self, obj):
            recordset = set()
            from_, below = obj.value
            for e, r in enumerate(self.records):
                f = r.get(obj.field)
                if f:
                    for v in f:
                        if v >= from_ and v < below:
                            recordset.add(e)
                            break
            return recordset

        def _above_to(self, obj):
            recordset = set()
            above, to = obj.value
            for e, r in enumerate(self.records):
                f = r.get(obj.field)
                if f:
                    for v in f:
                        if v > above and v <= to:
                            recordset.add(e)
                            break
            return recordset

        def _above_below(self, obj):
            recordset = set()
            above, below = obj.value
            for e, r in enumerate(self.records):
                f = r.get(obj.field)
                if f:
                    for v in f:
                        if v > above and v < below:
                            recordset.add(e)
                            break
            return recordset

        def _and(self, obj):
            return obj.left.result.answer.intersection(obj.result.answer)

        def _nor(self, obj):
            return obj.left.result.answer.intersection(
                self.existence - obj.result.answer)

        def _or(self, obj):
            return obj.left.result.answer.union(obj.result.answer)


    class Constraints(Processors):
        # Override the actions to evaluate non-index conditions.
        # This allows the constraints on record scans to be preserved for
        # comparison with expected values.
        # The constraints refer to objects which are modified in real actions.
            
        def non_index_condition(self, obj, record_number, record):
            pass
            
        def not_condition(self, obj):
            pass
            
        def operator(self, obj):
            pass
            
        def answer(self, obj):
            pass


    # validate() calls object.exists() to see if field <dbname> is in file
    # <dbset>.
    class Database():

        def exists(self, dbset, dbname):
            return dbname in dbset


    def setUp(self):
            
        self.processors = Where_validateTC.Processors()
        self.constraints = Where_validateTC.Constraints()
        self.database = Where_validateTC.Database()

    def tearDown(self):
        pass

    def test_validate_01(self):
        self.assertEqual(
            self._validate('f5 eq as and f5 like at or f2 is d'), [])
        self.assertEqual(
            self._validate('f5 ex d'), ['f5 ex d'])
        self.assertEqual(
            self._validate('f5 ex as and f5 like at or f2 is d'), ['f5 ex as'])
        self.assertEqual(
            self._validate('f5 eq as ant f5 like at or f2 is d'),
            ['eq', 'as ant f5'])
        self.assertEqual(
            self._validate('f5 eq as and (f2 is d or f5 like at)'), [])
        self.assertEqual(
            self._validate('f5 eq as if and (f2 is d or f5 like at)'), [])
        self.assertEqual(
            self._validate(''), [])
        self.assertEqual(
            self._validate(' '), [])
        self.assertEqual(
            self._validate('a'), ['a'])

    def _validate(self, query):
        w = where.Where(query)
        w.lex()
        w.parse()
        return w.validate(self.database, self.processors.file)
        

if __name__ == '__main__':
    runner = unittest.TextTestRunner
    loader = unittest.defaultTestLoader.loadTestsFromTestCase
    runner().run(loader(WhereTC))
    runner().run(loader(Where_MethodsTC))
    runner().run(loader(Where_errorTC))
    runner().run(loader(Where_lex_phraseTC))
    runner().run(loader(Where_parse_phraseTC))
    runner().run(loader(Where_parse_multi_phraseTC))
    runner().run(loader(Where_parse_not_phrasesTC))
    runner().run(loader(Where_parse_multi_not_phrasesTC))
    runner().run(loader(Where_parse_parenthesis_phrasesTC))
    runner().run(loader(Where_evaluate_traceTC))
    runner().run(loader(Where_evaluateTC))
    runner().run(loader(Where_validateTC))
